# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


import logging
from holado.common.context.session_context import SessionContext
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_core.common.tools.converters.converter import Converter
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_scope_manager():
    return __get_scenario_context().scope_manager

def __get_expression_evaluator():
    return __get_scenario_context().get_expression_evaluator()



#################################################
##  Steps to manage "if / else if /else"

@Given(r"if (?P<condition_expression>{Any}):", accept_condition=False)
def step_impl(context, condition_expression):
    __get_scope_manager().log_conditions_info(f"Before evaluation of if condition expression [{condition_expression}]")
    __get_scope_manager().increase_condition_level()
    _, cond = __get_expression_evaluator().evaluate_python_expression(condition_expression)
    if isinstance(cond, bool) or Converter.is_boolean(cond):
        if isinstance(cond, bool) and cond or Converter.is_boolean(cond) and Converter.to_boolean(cond):
            __get_scope_manager().enter_in_valid_condition()
    else:
        raise FunctionalException(f"Condition expression is not a boolean: [{condition_expression}] => [{cond}] (type: {Typing.get_object_class_fullname(cond)})")
    __get_scope_manager().log_conditions_info(f"After evaluation of if condition expression [{condition_expression}]")

@Given(r"else if (?P<condition_expression>{Any}):", accept_condition=False)
def step_impl(context, condition_expression):
    __get_scope_manager().leave_condition()
    if not __get_scope_manager().had_valid_condition():
        _, cond = __get_expression_evaluator().evaluate_python_expression(condition_expression)
        if isinstance(cond, bool) or Converter.is_boolean(cond):
            if isinstance(cond, bool) and cond or Converter.is_boolean(cond) and Converter.to_boolean(cond):
                __get_scope_manager().enter_in_valid_condition()
        else:
            raise FunctionalException(f"Condition expression is not a boolean: [{condition_expression}] => [{cond}] (type: {Typing.get_object_class_fullname(cond)})")
        __get_scope_manager().log_conditions_info(f"After evaluation of elif condition expression [{condition_expression}]")

@Given(r"else:", accept_condition=False)
def step_impl(context):
    __get_scope_manager().leave_condition()
    if not __get_scope_manager().had_valid_condition():
        __get_scope_manager().enter_in_valid_condition()
    __get_scope_manager().log_conditions_info(f"After else")

@Given(r"end if", accept_condition=False)
def step_impl(context):
    # TODO EKL: add verification that sub define is a 'if'
    __get_scope_manager().decrease_condition_level()









#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_rest.api.rest.rest_client import RestClient
from holado_python.common.tools.datetime import DateTime
import datetime
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)



class MockServerClient(RestClient):
    """
    MockServer client to its exposed REST API
    """
    
    @classmethod
    def is_available(cls):
        return RestClient.is_available()
    
    def __init__(self, name, url, headers=None):
        super().__init__(name, url, headers)
    
    def retrieve_request_responses(self, since_datetime=None, format_=None):
        """
        :param since_datetime: if defined, return only responses sent after this datetime
        :param format_: define response format (default: "json" ; supported values: "json", "log_entries")
        :returns: result of API 'retrieve' with type='request_responses'.
        """
        params = {'type': 'request_responses'}
        if format_ is not None:
            params['format'] = format_
            
        res = self.put('/mockserver/retrieve', params=params)

        if since_datetime is not None:
            if not isinstance(since_datetime, datetime.datetime):
                raise TechnicalException(f"Unmanaged since_datetime type '{Typing.get_object_class_fullname(since_datetime)}' (expected: datetime)")
            dt_str = DateTime.datetime_2_str(since_datetime, '%Y-%m-%d %H:%M:%S.%f')[:-3]
            res = [r for r in res if r['timestamp'] >= dt_str]
        
        return res
    
    def retrieve_pushed_data(self, since_datetime=None, method=None, body_type=None):
        result = self.retrieve_request_responses(since_datetime)
        
        res = []
        for request_response in result:
            if method is not None and request_response["httpRequest"]["method"] != method:
                continue
            if body_type is not None and isinstance(request_response["httpRequest"]["body"], dict) \
                    and "type" in request_response["httpRequest"]["body"] \
                    and request_response["httpRequest"]["body"]["type"] != body_type:
                continue
            
            if isinstance(request_response["httpRequest"]["body"], dict) and "type" in request_response["httpRequest"]["body"]:
                if "string" in request_response["httpRequest"]["body"]:
                    res.append(request_response["httpRequest"]["body"]["string"])
                else:
                    raise TechnicalException(f"Unmanaged type of body: {request_response['httpRequest']['body']}")
            else:
                res.append(request_response["httpRequest"]["body"])
        
        return res







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_ui_selenium.ui.gui.selenium.windows.selenium_window import SeleniumWindow
from holado_ui_selenium.ui.gui.selenium.inspectors.default_selenium_inspector import DefaultSeleniumInspector
from holado_ui_selenium.ui.gui.selenium.actors.default_selenium_actor import DefaultSeleniumActor

logger = logging.getLogger(__name__)



class DefaultSeleniumWindow(SeleniumWindow):
    """ Default Selenium window.
    """
    
    def __init__(self):
        super().__init__()
    
    def _initialize_inspect_builder(self):
        # # Get default (usually all modules and finder types)
        # res = self.inspector.inspect_builder
        #
        # # OR remove modules and add wanted with their default inspect builders
        # res = self.inspector.inspect_builder
        # res.remove_all_modules()
        # res.add_module("angular")
        # res.add_module("html")
        #
        # # OR remove a module and finder types from default inspect builders
        # res = self.inspector.inspect_builder
        # res.remove_module("angular")
        # res.get_inspect_builder_for_module("html").default_parameters.remove_finder_types([
        #         "text-node"
        #         ])
        #
        # # OR remove all finder types and add only wanted ones
        # res = self.inspector.inspect_builder
        # res.get_inspect_builder_for_module("html").default_parameters.remove_all_finder_types()
        # res.get_inspect_builder_for_module("html").default_parameters.add_finder_types([
        #         "text-node"
        #         ])
        
        # OR None to use default initialized in inspector
        res = None
        
        return res
        
    
    def _initialize_inspector(self):
        res = DefaultSeleniumInspector()
        res.initialize(self)
        return res

    def _initialize_actor(self):
        res = DefaultSeleniumActor()
        res.initialize(self)
        return res
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


#################################################
#
# Patches are done to follow ITU recommendation:
# https://www.e-navigation.nl/sites/default/files/R-REC-M.1371-5-201402-I!!PDF-E_1.pdf
#
#################################################


import logging
from bitarray import bitarray
import attr
import typing
import pyais
from pyais.messages import Payload, bit_field, from_mmsi, NMEA_VALUE, CommunicationStateMixin, from_speed, to_speed, \
    from_lat_lon, to_lat_lon, from_10th, to_10th, from_lat_lon_600, to_lat_lon_600, \
    from_turn, to_turn, ANY_MESSAGE, TagBlock, error_if_uninitialized, ASTERISK
from pyais.util import get_int, from_bytes_signed, from_bytes, bits2bytes, chunks, str_to_bin, bytes2bits, checksum
from pyais.exceptions import InvalidDataTypeException
from pyais.constants import NavigationStatus, EpfdType, ShipType, NavAid, StationType, TransmitMode, StationIntervals,\
    TurnRate, ManeuverIndicator
import functools

logger = logging.getLogger(__name__)



### Fix conversion of message types to/from bitarray

# All @ at end of string are considered padding.
# Note: in pyais implementation, decode is stopped at first @, which can be a regular field information, and following regular characters are skipped.
def HA_decode_bin_as_ascii6(bit_arr: bitarray) -> str:
    """
    Decode binary data as 6 bit ASCII.
    :param bit_arr: array of bits
    :return: ASCII String
    """
    string: str = ""
    c: bitarray
    for c in chunks(bit_arr, 6):  # type:ignore
        n: int = from_bytes(c.tobytes()) >> 2

        # Last entry may not have 6 bits
        if len(c) != 6:
            n >> (6 - len(c))

        if n < 0x20:
            n += 0x40

        string += chr(n)
    
    # Remove all @ at end since they correspond to padding
    string = string.rstrip('@')
    
    # Remove spaces as some encoders make padding with spaces instead of @
    return string.strip()

pyais.util.decode_bin_as_ascii6 = HA_decode_bin_as_ascii6


def HA_int_to_bin(val: typing.Union[int, bool], width: int, signed: bool = True) -> bitarray:
    """
    Convert an integer or boolean value to binary. 
    Compared to pyais implementation, the behaviour is changed if the value is too great to fit into
    `width` bits: 
      - in pyais, the maximum possible number that still fits is used,
      - in this implementation, a ValueError is raised.

    @param val:     Any integer or boolean value.
    @param width:   The bit width. If less than width bits are required, leading zeros are added.
    @param signed:  Set to True/False if the value is signed or not.
    @return:        The binary representation of value with exactly width bits. Type is bitarray.
    """
    # Compute the total number of bytes required to hold up to `width` bits.
    n_bytes, mod = divmod(width, 8)
    if mod > 0:
        n_bytes += 1

    # If the value is too big, return a bitarray of all 1's
    mask = (1 << width) - 1
    if val > mask:
        raise ValueError(f"Value {val} is too big for bit width {width} (max possible value: {mask})")

    bits = bitarray(endian='big')
    bits.frombytes(val.to_bytes(n_bytes, 'big', signed=signed))
    return bits[8 - mod if mod else 0:]

pyais.util.int_to_bin = HA_int_to_bin
pyais.messages.int_to_bin = HA_int_to_bin


@attr.s(slots=True)
class HAPayload(Payload):
    """Payload class with fix in to/from bitarray conversion.
    """

    @classmethod
    def from_bitarray(cls, bit_arr: bitarray) -> "ANY_MESSAGE":
        cur: int = 0
        end: int = 0
        length: int = len(bit_arr)
        kwargs: typing.Dict[str, typing.Any] = {}
        
        # Manage field with a variable length
        index_variable_length_field: int = None
        cursor_variable_length_field_beg: int = None
        cursor_variable_length_field_end: int = None

        # Iterate over the bits until a variable length field or the last bit of the bitarray or all fields are fully decoded
        for field_index, field in enumerate(cls.fields()):
            # Stop if field has a variable length
            if field.metadata['variable_length']:
                index_variable_length_field = field_index
                cursor_variable_length_field_beg = cur
                break
            
            if end >= length:
                # All fields that did not fit into the bit array are None
                kwargs[field.name] = None
                continue

            width = field.metadata['width']
            end = min(length, cur + width)
            bits = bit_arr[cur: end]
            
            kwargs[field.name] = cls._bitarray_to_field_type(bits, field)
            cur = end
        
        if index_variable_length_field is not None:
            # Iterate over fields in reverse order until variable length field
            end = length
            cur = length
            
            last_fields = tuple(cls.fields())[index_variable_length_field+1:]
            for field in reversed(last_fields):
                if end <= cursor_variable_length_field_beg:
                    # All fields that did not fit into the bit array are None
                    kwargs[field.name] = None
                    continue
                
                width = field.metadata['width']
                cur = max(cursor_variable_length_field_beg, end - width)
                bits = bit_arr[cur: end]
                
                kwargs[field.name] = cls._bitarray_to_field_type(bits, field)
                end = cur
            
            # Add variable length field
            cursor_variable_length_field_end = end
            field = cls.fields()[index_variable_length_field]
            bits = bit_arr[cursor_variable_length_field_beg: cursor_variable_length_field_end]
            kwargs[field.name] = cls._bitarray_to_field_type(bits, field)

        return cls(**kwargs)  # type:ignore

    @classmethod
    def _bitarray_to_field_type(cls, bits, field) -> typing.Any:
        val: typing.Any
        d_type = field.metadata['d_type']
        converter = field.metadata['to_converter']
        
        # Get the correct data type and decoding function
        if d_type == int or d_type == bool or d_type == float:
            shift = (8 - (len(bits) % 8)) % 8
            if field.metadata['signed']:
                val = from_bytes_signed(bits) >> shift
            else:
                val = from_bytes(bits) >> shift

            if d_type == float:
                val = float(val)
            elif d_type == bool:
                val = bool(val)

        elif d_type == str:
            val = HA_decode_bin_as_ascii6(bits)
        elif d_type == bytes:
            val = bits2bytes(bits)
        else:
            raise InvalidDataTypeException(d_type)

        val = converter(val) if converter is not None else val
        
        return val

    def to_bitarray(self) -> bitarray:
        """
        Convert all attributes of a given Payload/Message to binary.
        """
        out = bitarray()
        for field in self.fields():
            width = field.metadata['width']
            d_type = field.metadata['d_type']
            converter = field.metadata['from_converter']
            signed = field.metadata['signed']
            variable_length = field.metadata['variable_length']

            val = getattr(self, field.name)
            if val is None:
                continue

            val = converter(val) if converter is not None else val
            
            try:
                if d_type in (bool, int):
                    bits = HA_int_to_bin(val, width, signed=signed)
                elif d_type == float:
                    val = int(val)
                    bits = HA_int_to_bin(val, width, signed=signed)
                elif d_type == str:
                    trailing_spaces = not variable_length
                    bits = str_to_bin(val, width, trailing_spaces=trailing_spaces)
                elif d_type == bytes:
                    bits = bytes2bits(val, default=bitarray('0' * width))
                else:
                    raise InvalidDataTypeException(d_type)
            except ValueError as exc:
                raise ValueError(f"Invalid value for field '{field.name}': {exc}") from None

            bits = bits[:width]
            out += bits

        return out


# Fix message types:
# - Use fixes of Payload implemented in HAPayload, in all message types
# - type 21: exclude full_name property from asdict result
# - type 26: fix create and from_bitarray methods

@attr.s(slots=True)
class HAMessageType1(HAPayload, CommunicationStateMixin):
    """
    AIS Vessel position report using SOTDMA (Self-Organizing Time Division Multiple Access)
    """
    msg_type = bit_field(6, int, default=1, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    status = bit_field(4, int, default=0, converter=NavigationStatus.from_value, signed=False)
    turn = bit_field(8, float, default=TurnRate.NO_TI_DEFAULT, signed=True, to_converter=to_turn, from_converter=from_turn)
    speed = bit_field(10, float, from_converter=from_speed, to_converter=to_speed, default=0, signed=False)
    accuracy = bit_field(1, bool, default=0, signed=False)
    lon = bit_field(28, float, from_converter=from_lat_lon, to_converter=to_lat_lon, default=0, signed=True)
    lat = bit_field(27, float, from_converter=from_lat_lon, to_converter=to_lat_lon, default=0, signed=True)
    course = bit_field(12, float, from_converter=from_10th, to_converter=to_10th, default=0, signed=False)
    heading = bit_field(9, int, default=0, signed=False)
    second = bit_field(6, int, default=0, signed=False)
    maneuver = bit_field(2, int, default=0, from_converter=ManeuverIndicator.from_value,
                         to_converter=ManeuverIndicator.from_value, signed=False)
    spare_1 = bit_field(3, bytes, default=b'', is_spare=True)
    raim = bit_field(1, bool, default=0)
    radio = bit_field(19, int, default=0, signed=False)

pyais.messages.MessageType1 = HAMessageType1
pyais.messages.MSG_CLASS[1] = pyais.messages.MessageType1


class HAMessageType2(HAMessageType1):
    """
    AIS Vessel position report using SOTDMA (Self-Organizing Time Division Multiple Access)
    """
    msg_type = bit_field(6, int, default=2)

pyais.messages.MessageType2 = HAMessageType2
pyais.messages.MSG_CLASS[2] = pyais.messages.MessageType2


class HAMessageType3(HAMessageType1):
    """
    AIS Vessel position report using ITDMA (Incremental Time Division Multiple Access)
    """
    msg_type = bit_field(6, int, default=3)

pyais.messages.MessageType3 = HAMessageType3
pyais.messages.MSG_CLASS[3] = pyais.messages.MessageType3


@attr.s(slots=True)
class HAMessageType4(HAPayload, CommunicationStateMixin):
    """
    AIS Vessel position report using SOTDMA (Self-Organizing Time Division Multiple Access)
    """
    msg_type = bit_field(6, int, default=4, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    year = bit_field(14, int, default=1970, signed=False)
    month = bit_field(4, int, default=1, signed=False)
    day = bit_field(5, int, default=1, signed=False)
    hour = bit_field(5, int, default=0, signed=False)
    minute = bit_field(6, int, default=0, signed=False)
    second = bit_field(6, int, default=0, signed=False)
    accuracy = bit_field(1, bool, default=0, signed=False)
    lon = bit_field(28, float, from_converter=from_lat_lon, to_converter=to_lat_lon, signed=True, default=0)
    lat = bit_field(27, float, from_converter=from_lat_lon, to_converter=to_lat_lon, signed=True, default=0)
    epfd = bit_field(4, int, default=0, from_converter=EpfdType.from_value, to_converter=EpfdType.from_value,
                     signed=False)
    spare_1 = bit_field(10, bytes, default=b'', is_spare=True)
    raim = bit_field(1, bool, default=0)
    radio = bit_field(19, int, default=0, signed=False)

pyais.messages.MessageType4 = HAMessageType4
pyais.messages.MSG_CLASS[4] = pyais.messages.MessageType4


@attr.s(slots=True)
class HAMessageType5(HAPayload):
    """
    Static and Voyage Related Data
    """
    msg_type = bit_field(6, int, default=5, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    ais_version = bit_field(2, int, default=0, signed=False)
    imo = bit_field(30, int, default=0, signed=False)
    callsign = bit_field(42, str, default='')
    shipname = bit_field(120, str, default='')
    ship_type = bit_field(8, int, default=0, from_converter=ShipType.from_value, to_converter=ShipType.from_value)
    to_bow = bit_field(9, int, default=0, signed=False)
    to_stern = bit_field(9, int, default=0, signed=False)
    to_port = bit_field(6, int, default=0, signed=False)
    to_starboard = bit_field(6, int, default=0, signed=False)
    epfd = bit_field(4, int, default=0, from_converter=EpfdType.from_value, to_converter=EpfdType.from_value)
    month = bit_field(4, int, default=0, signed=False)
    day = bit_field(5, int, default=0, signed=False)
    hour = bit_field(5, int, default=0, signed=False)
    minute = bit_field(6, int, default=0, signed=False)
    draught = bit_field(8, float, from_converter=from_10th, to_converter=to_10th, default=0, signed=False)
    destination = bit_field(120, str, default='')
    dte = bit_field(1, bool, default=0, signed=False)
    spare_1 = bit_field(1, bytes, default=b'', is_spare=True)

pyais.messages.MessageType5 = HAMessageType5
pyais.messages.MSG_CLASS[5] = pyais.messages.MessageType5


@attr.s(slots=True)
class HAMessageType6(HAPayload):
    """
    Binary Addresses Message
    """
    msg_type = bit_field(6, int, default=6)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    seqno = bit_field(2, int, default=0, signed=False)
    dest_mmsi = bit_field(30, int, from_converter=from_mmsi)
    retransmit = bit_field(1, bool, default=False, signed=False)
    spare_1 = bit_field(1, bytes, default=b'', is_spare=True)
    dac = bit_field(10, int, default=0, signed=False)
    fid = bit_field(6, int, default=0, signed=False)
    data = bit_field(920, bytes, default=b'', variable_length=True)

pyais.messages.MessageType6 = HAMessageType6
pyais.messages.MSG_CLASS[6] = pyais.messages.MessageType6


@attr.s(slots=True)
class HAMessageType7(HAPayload):
    """
    Binary Acknowledge
    """
    msg_type = bit_field(6, int, default=7, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)
    mmsi1 = bit_field(30, int, default=0, from_converter=from_mmsi)
    mmsiseq1 = bit_field(2, int, default=0, signed=False)
    mmsi2 = bit_field(30, int, default=0, from_converter=from_mmsi)
    mmsiseq2 = bit_field(2, int, default=0, signed=False)
    mmsi3 = bit_field(30, int, default=0, from_converter=from_mmsi)
    mmsiseq3 = bit_field(2, int, default=0, signed=False)
    mmsi4 = bit_field(30, int, default=0, from_converter=from_mmsi)
    mmsiseq4 = bit_field(2, int, default=0, signed=False)

pyais.messages.MessageType7 = HAMessageType7
pyais.messages.MSG_CLASS[7] = pyais.messages.MessageType7


@attr.s(slots=True)
class HAMessageType8(HAPayload):
    """
    Binary Acknowledge
    """
    msg_type = bit_field(6, int, default=8, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)
    dac = bit_field(10, int, default=0, signed=False)
    fid = bit_field(6, int, default=0, signed=False)
    data = bit_field(952, bytes, default=b'', variable_length=True)

pyais.messages.MessageType8 = HAMessageType8
pyais.messages.MSG_CLASS[8] = pyais.messages.MessageType8


@attr.s(slots=True)
class HAMessageType9(HAPayload, CommunicationStateMixin):
    """
    Standard SAR Aircraft Position Report
    """
    msg_type = bit_field(6, int, default=9, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    alt = bit_field(12, int, default=0, signed=False)
    # speed over ground is in knots, not deciknots
    speed = bit_field(10, float, default=0, signed=False)
    accuracy = bit_field(1, bool, default=0, signed=False)
    lon = bit_field(28, float, from_converter=from_lat_lon, to_converter=to_lat_lon, signed=True, default=0)
    lat = bit_field(27, float, from_converter=from_lat_lon, to_converter=to_lat_lon, signed=True, default=0)
    course = bit_field(12, float, from_converter=from_10th, to_converter=to_10th, default=0, signed=False)
    second = bit_field(6, int, default=0, signed=False)

    reserved_1 = bit_field(8, int, default=0)
    dte = bit_field(1, bool, default=0)
    spare_1 = bit_field(3, bytes, default=b'', is_spare=True)
    assigned = bit_field(1, bool, default=0)
    raim = bit_field(1, bool, default=0)
    radio = bit_field(20, int, default=0, signed=False)

pyais.messages.MessageType9 = HAMessageType9
pyais.messages.MSG_CLASS[9] = pyais.messages.MessageType9


@attr.s(slots=True)
class HAMessageType10(HAPayload):
    """
    UTC/Date Inquiry
    """
    msg_type = bit_field(6, int, default=10, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)
    dest_mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_2 = bit_field(2, bytes, default=b'', is_spare=True)

pyais.messages.MessageType10 = HAMessageType10
pyais.messages.MSG_CLASS[10] = pyais.messages.MessageType10


class HAMessageType11(HAMessageType4):
    """
    UTC/Date Response
    """
    msg_type = bit_field(6, int, default=11, signed=False)

pyais.messages.MessageType11 = HAMessageType11
pyais.messages.MSG_CLASS[11] = pyais.messages.MessageType11


@attr.s(slots=True)
class HAMessageType12(HAPayload):
    """
    Addressed Safety-Related Message
    """
    msg_type = bit_field(6, int, default=12, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    seqno = bit_field(2, int, default=0, signed=False)
    dest_mmsi = bit_field(30, int, from_converter=from_mmsi)
    retransmit = bit_field(1, bool, default=False, signed=False)
    spare_1 = bit_field(1, bytes, default=b'', is_spare=True)
    text = bit_field(936, str, default='', variable_length=True)

pyais.messages.MessageType12 = HAMessageType12
pyais.messages.MSG_CLASS[12] = pyais.messages.MessageType12


class HAMessageType13(HAMessageType7):
    """
    Identical to type 7
    """
    msg_type = bit_field(6, int, default=13, signed=False)

pyais.messages.MessageType13 = HAMessageType13
pyais.messages.MSG_CLASS[13] = pyais.messages.MessageType13


@attr.s(slots=True)
class HAMessageType14(HAPayload):
    """
    Safety-Related Broadcast Message
    """
    msg_type = bit_field(6, int, default=14, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)
    text = bit_field(968, str, default='', variable_length=True)

pyais.messages.MessageType14 = HAMessageType14
pyais.messages.MSG_CLASS[14] = pyais.messages.MessageType14


@attr.s(slots=True)
class HAMessageType15(HAPayload):
    """
    Interrogation
    """
    msg_type = bit_field(6, int, default=15, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)
    mmsi1 = bit_field(30, int, default=0, from_converter=from_mmsi)
    type1_1 = bit_field(6, int, default=0, signed=False)
    offset1_1 = bit_field(12, int, default=0, signed=False)
    spare_2 = bit_field(2, bytes, default=b'', is_spare=True)
    type1_2 = bit_field(6, int, default=0, signed=False)
    offset1_2 = bit_field(12, int, default=0, signed=False)
    spare_3 = bit_field(2, bytes, default=b'', is_spare=True)
    mmsi2 = bit_field(30, int, default=0, from_converter=from_mmsi)
    type2_1 = bit_field(6, int, default=0, signed=False)
    offset2_1 = bit_field(12, int, default=0, signed=False)
    spare_4 = bit_field(2, bytes, default=b'', is_spare=True)

pyais.messages.MessageType15 = HAMessageType15
pyais.messages.MSG_CLASS[15] = pyais.messages.MessageType15


@attr.s(slots=True)
class HAMessageType16DestinationA(HAPayload):
    """
    Assignment Mode Command
    """
    msg_type = bit_field(6, int, default=16, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)

    mmsi1 = bit_field(30, int, default=0, from_converter=from_mmsi)
    offset1 = bit_field(12, int, default=0, signed=False)
    increment1 = bit_field(10, int, default=0, signed=False)
    spare_2 = bit_field(4, bytes, default=b'', is_spare=True)

pyais.messages.MessageType16DestinationA = HAMessageType16DestinationA


@attr.s(slots=True)
class HAMessageType16DestinationAB(HAPayload):
    """
    Assignment Mode Command
    """
    msg_type = bit_field(6, int, default=16, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)

    mmsi1 = bit_field(30, int, default=0, from_converter=from_mmsi)
    offset1 = bit_field(12, int, default=0, signed=False)
    increment1 = bit_field(10, int, default=0, signed=False)

    mmsi2 = bit_field(30, int, default=0, from_converter=from_mmsi)
    offset2 = bit_field(12, int, default=0, signed=False)
    increment2 = bit_field(10, int, default=0, signed=False)

pyais.messages.MessageType16DestinationAB = HAMessageType16DestinationAB


@attr.s(slots=True)
class HAMessageType16(HAPayload):
    @classmethod
    def create(cls, **kwargs: typing.Union[str, float, int, bool, bytes]) -> "ANY_MESSAGE":
        if 'mmsi2' in kwargs:
            return HAMessageType16DestinationAB.create(**kwargs)
        else:
            return HAMessageType16DestinationA.create(**kwargs)

    @classmethod
    def from_bitarray(cls, bit_arr: bitarray) -> "ANY_MESSAGE":
        if len(bit_arr) > 96:
            return HAMessageType16DestinationAB.from_bitarray(bit_arr)
        else:
            return HAMessageType16DestinationA.from_bitarray(bit_arr)

pyais.messages.MessageType16 = HAMessageType16
pyais.messages.MSG_CLASS[16] = pyais.messages.MessageType16


@attr.s(slots=True)
class HAMessageType17(HAPayload):
    """
    DGNSS Broadcast Binary Message
    """
    msg_type = bit_field(6, int, default=17, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)
    # Note that latitude and longitude are in units of a tenth of a minute
    lon = bit_field(18, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)
    lat = bit_field(17, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)
    spare_2 = bit_field(5, bytes, default=b'', is_spare=True)
    data = bit_field(736, bytes, default=b'', variable_length=True)

pyais.messages.MessageType17 = HAMessageType17
pyais.messages.MSG_CLASS[17] = pyais.messages.MessageType17


@attr.s(slots=True)
class HAMessageType18(HAPayload, CommunicationStateMixin):
    """
    Standard Class B CS Position Report
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_18_standard_class_b_cs_position_report
    """
    msg_type = bit_field(6, int, default=18, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    reserved_1 = bit_field(8, int, default=0, signed=False)
    speed = bit_field(10, float, from_converter=from_speed, to_converter=to_speed, default=0, signed=False)
    accuracy = bit_field(1, bool, default=0, signed=False)
    lon = bit_field(28, float, from_converter=from_lat_lon, to_converter=to_lat_lon, signed=True, default=0)
    lat = bit_field(27, float, from_converter=from_lat_lon, to_converter=to_lat_lon, signed=True, default=0)
    course = bit_field(12, float, from_converter=from_10th, to_converter=to_10th, default=0, signed=False)
    heading = bit_field(9, int, default=0, signed=False)
    second = bit_field(6, int, default=0, signed=False)
    reserved_2 = bit_field(2, int, default=0, signed=False)
    cs = bit_field(1, bool, default=0, signed=False)
    display = bit_field(1, bool, default=0)
    dsc = bit_field(1, bool, default=0)
    band = bit_field(1, bool, default=0)
    msg22 = bit_field(1, bool, default=0)
    assigned = bit_field(1, bool, default=0)
    raim = bit_field(1, bool, default=0)
    radio = bit_field(20, int, default=0)

pyais.messages.MessageType18 = HAMessageType18
pyais.messages.MSG_CLASS[18] = pyais.messages.MessageType18


@attr.s(slots=True)
class HAMessageType19(HAPayload):
    """
    Extended Class B CS Position Report
    """
    msg_type = bit_field(6, int, default=19, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    reserved_1 = bit_field(8, int, default=0)

    speed = bit_field(10, float, from_converter=from_speed, to_converter=to_speed, default=0, signed=False)
    accuracy = bit_field(1, bool, default=0, signed=False)
    lon = bit_field(28, float, from_converter=from_lat_lon, to_converter=to_lat_lon, signed=True, default=0)
    lat = bit_field(27, float, from_converter=from_lat_lon, to_converter=to_lat_lon, signed=True, default=0)
    course = bit_field(12, float, from_converter=from_10th, to_converter=to_10th, default=0, signed=False)
    heading = bit_field(9, int, default=0, signed=False)
    second = bit_field(6, int, default=0, signed=False)
    reserved_2 = bit_field(4, int, default=0, signed=False)
    shipname = bit_field(120, str, default='')
    ship_type = bit_field(8, int, default=0, from_converter=ShipType.from_value, to_converter=ShipType.from_value,
                          signed=False)
    to_bow = bit_field(9, int, default=0, signed=False)
    to_stern = bit_field(9, int, default=0, signed=False)
    to_port = bit_field(6, int, default=0, signed=False)
    to_starboard = bit_field(6, int, default=0, signed=False)
    epfd = bit_field(4, int, default=0, from_converter=EpfdType.from_value, to_converter=EpfdType.from_value)
    raim = bit_field(1, bool, default=0)
    dte = bit_field(1, bool, default=0)
    assigned = bit_field(1, bool, default=0, signed=False)
    spare_1 = bit_field(4, bytes, default=b'', is_spare=True)

pyais.messages.MessageType19 = HAMessageType19
pyais.messages.MSG_CLASS[19] = pyais.messages.MessageType19


@attr.s(slots=True)
class HAMessageType20ReservationBlock1(HAPayload):
    """
    Data Link Management Message
    """
    msg_type = bit_field(6, int, default=20, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)

    offset1 = bit_field(12, int, default=0, signed=False)
    number1 = bit_field(4, int, default=0, signed=False)
    timeout1 = bit_field(3, int, default=0, signed=False)
    increment1 = bit_field(11, int, default=0, signed=False)
    spare_2 = bit_field(2, bytes, default=b'', is_spare=True)

pyais.messages.MessageType20ReservationBlock1 = HAMessageType20ReservationBlock1


@attr.s(slots=True)
class HAMessageType20ReservationBlock12(HAPayload):
    """
    Data Link Management Message
    """
    msg_type = bit_field(6, int, default=20, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)

    offset1 = bit_field(12, int, default=0, signed=False)
    number1 = bit_field(4, int, default=0, signed=False)
    timeout1 = bit_field(3, int, default=0, signed=False)
    increment1 = bit_field(11, int, default=0, signed=False)

    offset2 = bit_field(12, int, default=0, signed=False)
    number2 = bit_field(4, int, default=0, signed=False)
    timeout2 = bit_field(3, int, default=0, signed=False)
    increment2 = bit_field(11, int, default=0, signed=False)
    spare_2 = bit_field(4, bytes, default=b'', is_spare=True)

pyais.messages.MessageType20ReservationBlock12 = HAMessageType20ReservationBlock12


@attr.s(slots=True)
class HAMessageType20ReservationBlock123(HAPayload):
    """
    Data Link Management Message
    """
    msg_type = bit_field(6, int, default=20, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)

    offset1 = bit_field(12, int, default=0, signed=False)
    number1 = bit_field(4, int, default=0, signed=False)
    timeout1 = bit_field(3, int, default=0, signed=False)
    increment1 = bit_field(11, int, default=0, signed=False)

    offset2 = bit_field(12, int, default=0, signed=False)
    number2 = bit_field(4, int, default=0, signed=False)
    timeout2 = bit_field(3, int, default=0, signed=False)
    increment2 = bit_field(11, int, default=0, signed=False)

    offset3 = bit_field(12, int, default=0, signed=False)
    number3 = bit_field(4, int, default=0, signed=False)
    timeout3 = bit_field(3, int, default=0, signed=False)
    increment3 = bit_field(11, int, default=0, signed=False)
    spare_2 = bit_field(6, bytes, default=b'', is_spare=True)

pyais.messages.MessageType20ReservationBlock123 = HAMessageType20ReservationBlock123


@attr.s(slots=True)
class HAMessageType20ReservationBlock1234(HAPayload):
    """
    Data Link Management Message
    """
    msg_type = bit_field(6, int, default=20, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)

    offset1 = bit_field(12, int, default=0, signed=False)
    number1 = bit_field(4, int, default=0, signed=False)
    timeout1 = bit_field(3, int, default=0, signed=False)
    increment1 = bit_field(11, int, default=0, signed=False)

    offset2 = bit_field(12, int, default=0, signed=False)
    number2 = bit_field(4, int, default=0, signed=False)
    timeout2 = bit_field(3, int, default=0, signed=False)
    increment2 = bit_field(11, int, default=0, signed=False)

    offset3 = bit_field(12, int, default=0, signed=False)
    number3 = bit_field(4, int, default=0, signed=False)
    timeout3 = bit_field(3, int, default=0, signed=False)
    increment3 = bit_field(11, int, default=0, signed=False)

    offset4 = bit_field(12, int, default=0, signed=False)
    number4 = bit_field(4, int, default=0, signed=False)
    timeout4 = bit_field(3, int, default=0, signed=False)
    increment4 = bit_field(11, int, default=0, signed=False)

pyais.messages.MessageType20ReservationBlock1234 = HAMessageType20ReservationBlock1234


@attr.s(slots=True)
class HAMessageType20(HAPayload):
    @classmethod
    def create(cls, **kwargs: typing.Union[str, float, int, bool, bytes]) -> "ANY_MESSAGE":
        if 'offset4' in kwargs and int(kwargs['offset4']) > 0:
            return HAMessageType20ReservationBlock1234.create(**kwargs)
        elif 'offset3' in kwargs and int(kwargs['offset3']) > 0:
            return HAMessageType20ReservationBlock123.create(**kwargs)
        elif 'offset2' in kwargs and int(kwargs['offset2']) > 0:
            return HAMessageType20ReservationBlock12.create(**kwargs)
        else:
            return HAMessageType20ReservationBlock1.create(**kwargs)

    @classmethod
    def from_bitarray(cls, bit_arr: bitarray) -> "ANY_MESSAGE":
        if len(bit_arr) > 136:
            return HAMessageType20ReservationBlock1234.from_bitarray(bit_arr)
        elif len(bit_arr) > 104:
            return HAMessageType20ReservationBlock123.from_bitarray(bit_arr)
        elif len(bit_arr) > 72:
            return HAMessageType20ReservationBlock12.from_bitarray(bit_arr)
        else:
            return HAMessageType20ReservationBlock1.from_bitarray(bit_arr)

pyais.messages.MessageType20 = HAMessageType20
pyais.messages.MSG_CLASS[20] = pyais.messages.MessageType20


@attr.s(slots=True)
class HAMessageType21(HAPayload):
    """
    Aid-to-Navigation Report
    """
    msg_type = bit_field(6, int, default=21, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    aid_type = bit_field(5, int, default=0, from_converter=NavAid.from_value, to_converter=NavAid.from_value,
                         signed=False)
    name = bit_field(120, str, default='')

    accuracy = bit_field(1, bool, default=0, signed=False)
    lon = bit_field(28, float, from_converter=from_lat_lon, to_converter=to_lat_lon, signed=True, default=0)
    lat = bit_field(27, float, from_converter=from_lat_lon, to_converter=to_lat_lon, signed=True, default=0)
    to_bow = bit_field(9, int, default=0, signed=False)
    to_stern = bit_field(9, int, default=0, signed=False)
    to_port = bit_field(6, int, default=0, signed=False)
    to_starboard = bit_field(6, int, default=0, signed=False)

    epfd = bit_field(4, int, default=0, from_converter=EpfdType.from_value, to_converter=EpfdType.from_value)
    second = bit_field(6, int, default=0, signed=False)
    off_position = bit_field(1, bool, default=0)
    reserved_1 = bit_field(8, int, default=0, signed=False)
    raim = bit_field(1, bool, default=0)
    virtual_aid = bit_field(1, bool, default=0)
    assigned = bit_field(1, bool, default=0)
    spare_1 = bit_field(1, bytes, default=b'', is_spare=True)
    name_ext = bit_field(88, str, default='')

    @functools.cached_property
    def full_name(self) -> str:
        """The name field is up to 20 characters of 6-bit ASCII. If this field
        is full (has no trailing @ characters) the decoder should interpret
        the Name Extension field later in the message (no more than 14 6-bit
        characters) and concatenate it to this one to obtain the full name."""
        if self.name:
            if self.name_ext:
                return f"{self.name}{self.name_ext}"
            return str(self.name)
        return ""
    
    # TODO: replace following override by a generic implementation in Payload.asdict that exports only fields
    def asdict(self, enum_as_int: bool = False) -> typing.Dict[str, typing.Optional[NMEA_VALUE]]:
        res = super().asdict(enum_as_int)
        
        # Remove key 'full_name' if present
        if 'full_name' in res:
            del res['full_name']
        
        return res

pyais.messages.MessageType21 = HAMessageType21
pyais.messages.MSG_CLASS[21] = pyais.messages.MessageType21


@attr.s(slots=True)
class HAMessageType22Addressed(HAPayload):
    """
    Channel Management
    """
    msg_type = bit_field(6, int, default=22, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)

    channel_a = bit_field(12, int, default=0, signed=False)
    channel_b = bit_field(12, int, default=0, signed=False)
    txrx = bit_field(4, int, default=0, signed=False)
    power = bit_field(1, bool, default=0)  # 69 bits

    # If it is addressed (addressed field is 1),
    # the same span of data is interpreted as two 30-bit MMSIs
    # beginning at bit offsets 69 and 104 respectively.
    dest1 = bit_field(30, int, default=0, from_converter=from_mmsi)
    empty_1 = bit_field(5, int, default=0)
    dest2 = bit_field(30, int, default=0, from_converter=from_mmsi)
    empty_2 = bit_field(5, int, default=0)

    addressed = bit_field(1, bool, default=0)
    band_a = bit_field(1, bool, default=0)
    band_b = bit_field(1, bool, default=0)
    zonesize = bit_field(3, int, default=0)
    spare_2 = bit_field(23, bytes, default=b'', is_spare=True)

pyais.messages.MessageType22Addressed = HAMessageType22Addressed


@attr.s(slots=True)
class HAMessageType22Broadcast(HAPayload):
    """
    Channel Management
    """
    msg_type = bit_field(6, int, default=22, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)

    channel_a = bit_field(12, int, default=0, signed=False)
    channel_b = bit_field(12, int, default=0, signed=False)
    txrx = bit_field(4, int, default=0, signed=False)
    power = bit_field(1, bool, default=0)

    # If the message is broadcast (addressed field is 0),
    # the ne_lon, ne_lat, sw_lon, and sw_lat fields are the
    # corners of a rectangular jurisdiction area over which control parameter
    # ne_lon, ne_lat, sw_lon, and sw_lat fields are in 0.1 minutes
    ne_lon = bit_field(18, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)
    ne_lat = bit_field(17, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)
    sw_lon = bit_field(18, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)
    sw_lat = bit_field(17, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)

    addressed = bit_field(1, bool, default=0)
    band_a = bit_field(1, bool, default=0)
    band_b = bit_field(1, bool, default=0)
    zonesize = bit_field(3, int, default=0, signed=False)
    spare_2 = bit_field(23, bytes, default=b'', is_spare=True)

pyais.messages.MessageType22Broadcast = HAMessageType22Broadcast


@attr.s(slots=True)
class HAMessageType23(HAPayload):
    """
    Group Assignment Command
    """
    msg_type = bit_field(6, int, default=23, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)
    spare_1 = bit_field(2, bytes, default=b'', is_spare=True)

    ne_lon = bit_field(18, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)
    ne_lat = bit_field(17, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)
    sw_lon = bit_field(18, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)
    sw_lat = bit_field(17, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)

    station_type = bit_field(4, int, default=0, from_converter=StationType.from_value,
                             to_converter=StationType.from_value)
    ship_type = bit_field(8, int, default=0, from_converter=ShipType.from_value, to_converter=ShipType.from_value)
    spare_2 = bit_field(22, bytes, default=b'', is_spare=True)

    txrx = bit_field(2, int, default=0, from_converter=TransmitMode.from_value, to_converter=TransmitMode.from_value,
                     signed=False)
    interval = bit_field(4, int, default=0, from_converter=StationIntervals.from_value,
                         to_converter=StationIntervals.from_value)
    quiet = bit_field(4, int, default=0, signed=False)
    spare_3 = bit_field(6, bytes, default=b'', is_spare=True)

pyais.messages.MessageType23 = HAMessageType23
pyais.messages.MSG_CLASS[23] = pyais.messages.MessageType23


@attr.s(slots=True)
class HAMessageType24PartA(HAPayload):
    msg_type = bit_field(6, int, default=24, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    # partno = bit_field(2, int, default=0, signed=False)
    reserved = bit_field(1, int, default=0, signed=False, is_spare=True)
    partno = bit_field(1, int, default=0, signed=False)
    shipname = bit_field(120, str, default='')
    spare_1 = bit_field(8, bytes, default=b'', is_spare=True)

pyais.messages.MessageType24PartA = HAMessageType24PartA


@attr.s(slots=True)
class HAMessageType24PartB(HAPayload):
    msg_type = bit_field(6, int, default=24, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    # partno = bit_field(2, int, default=0, signed=False)
    reserved = bit_field(1, int, default=0, signed=False, is_spare=True)
    partno = bit_field(1, int, default=0, signed=False)
    ship_type = bit_field(8, int, default=0, signed=False)
    vendorid = bit_field(18, str, default='', signed=False)
    model = bit_field(4, int, default=0, signed=False)
    serial = bit_field(20, int, default=0, signed=False)
    callsign = bit_field(42, str, default='')

    to_bow = bit_field(9, int, default=0, signed=False)
    to_stern = bit_field(9, int, default=0, signed=False)
    to_port = bit_field(6, int, default=0, signed=False)
    to_starboard = bit_field(6, int, default=0, signed=False)

    spare_1 = bit_field(6, bytes, default=b'', is_spare=True)

pyais.messages.MessageType24PartB = HAMessageType24PartB


@attr.s(slots=True)
class HAMessageType25AddressedStructured(HAPayload):
    msg_type = bit_field(6, int, default=25, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    addressed = bit_field(1, bool, default=0, signed=False)
    structured = bit_field(1, bool, default=0, signed=False)

    dest_mmsi = bit_field(30, int, default=0, from_converter=from_mmsi, signed=False)
    spare = bit_field(2, int, default=0, signed=False, is_spare=True)
    app_id = bit_field(16, int, default=0, signed=False)
    data = bit_field(80, bytes, default=b'', variable_length=True)
    
pyais.messages.MessageType25AddressedStructured = HAMessageType25AddressedStructured


@attr.s(slots=True)
class HAMessageType25AddressedUnstructured(HAPayload):
    msg_type = bit_field(6, int, default=25, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    addressed = bit_field(1, bool, default=0, signed=False)
    structured = bit_field(1, bool, default=0, signed=False)

    dest_mmsi = bit_field(30, int, default=0, from_converter=from_mmsi)
    spare = bit_field(2, int, default=0, signed=False, is_spare=True)
    data = bit_field(96, bytes, default=b'', variable_length=True)

pyais.messages.MessageType25AddressedUnstructured = HAMessageType25AddressedUnstructured


@attr.s(slots=True)
class HAMessageType26AddressedStructured(HAPayload, CommunicationStateMixin):
    msg_type = bit_field(6, int, default=26, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    addressed = bit_field(1, bool, default=0, signed=False)
    structured = bit_field(1, bool, default=0, signed=False)

    dest_mmsi = bit_field(30, int, default=0, from_converter=from_mmsi)
    spare1 = bit_field(2, int, default=0, signed=False, is_spare=True)
    app_id = bit_field(16, int, default=0, signed=False)
    data = bit_field(952, bytes, default=b'', variable_length=True)
    spare2 = bit_field(4, int, default=0, signed=False, is_spare=True)
    radio = bit_field(20, int, default=0, signed=False)

pyais.messages.MessageType26AddressedStructured = HAMessageType26AddressedStructured


@attr.s(slots=True)
class HAMessageType26BroadcastStructured(HAPayload, CommunicationStateMixin):
    msg_type = bit_field(6, int, default=26, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    addressed = bit_field(1, bool, default=0, signed=False)
    structured = bit_field(1, bool, default=0, signed=False)

    app_id = bit_field(16, int, default=0, signed=False)
    data = bit_field(984, bytes, default=b'', variable_length=True)
    spare2 = bit_field(4, int, default=0, signed=False, is_spare=True)
    radio = bit_field(20, int, default=0, signed=False)

pyais.messages.MessageType26BroadcastStructured = HAMessageType26BroadcastStructured


@attr.s(slots=True)
class HAMessageType26AddressedUnstructured(HAPayload, CommunicationStateMixin):
    msg_type = bit_field(6, int, default=26, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    addressed = bit_field(1, bool, default=0, signed=False)
    structured = bit_field(1, bool, default=0, signed=False)

    dest_mmsi = bit_field(30, int, default=0, from_converter=from_mmsi)
    spare1 = bit_field(2, int, default=0, signed=False, is_spare=True)
    data = bit_field(968, bytes, default=b'', variable_length=True)
    spare2 = bit_field(4, int, default=0, signed=False, is_spare=True)
    radio = bit_field(20, int, default=0, signed=False)

pyais.messages.MessageType26AddressedUnstructured = HAMessageType26AddressedUnstructured


@attr.s(slots=True)
class HAMessageType26BroadcastUnstructured(HAPayload, CommunicationStateMixin):
    msg_type = bit_field(6, int, default=26, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    addressed = bit_field(1, bool, default=0, signed=False)
    structured = bit_field(1, bool, default=0, signed=False)

    data = bit_field(1000, bytes, default=b'', variable_length=True)
    spare2 = bit_field(4, int, default=0, signed=False, is_spare=True)
    radio = bit_field(20, int, default=0, signed=False)

pyais.messages.MessageType26BroadcastUnstructured = HAMessageType26BroadcastUnstructured


class HAMessageType26(HAPayload):
    """
    Multiple Slot Binary Message

    NOTE: This message type is quite uncommon and
    I was not able find any real world occurrence of the type.
    Also documentation seems to vary. Use with caution.
    """

    @classmethod
    def create(cls, **kwargs: typing.Union[str, float, int, bool, bytes]) -> "ANY_MESSAGE":
        addressed = kwargs.get('addressed', False)
        structured = kwargs.get('structured', False)

        if addressed:
            if structured:
                return HAMessageType26AddressedStructured.create(**kwargs)
            else:
                return HAMessageType26AddressedUnstructured.create(**kwargs)
        else:
            if structured:
                return HAMessageType26BroadcastStructured.create(**kwargs)
            else:
                return HAMessageType26BroadcastUnstructured.create(**kwargs)

    @classmethod
    def from_bitarray(cls, bit_arr: bitarray) -> "ANY_MESSAGE":
        addressed: int = get_int(bit_arr, 38, 39)
        structured: int = get_int(bit_arr, 39, 40)

        if addressed:
            if structured:
                return HAMessageType26AddressedStructured.from_bitarray(bit_arr)
            else:
                return HAMessageType26AddressedUnstructured.from_bitarray(bit_arr)
        else:
            if structured:
                return HAMessageType26BroadcastStructured.from_bitarray(bit_arr)
            else:
                return HAMessageType26BroadcastUnstructured.from_bitarray(bit_arr)

pyais.messages.MessageType26 = HAMessageType26
pyais.messages.MSG_CLASS[26] = pyais.messages.MessageType26


@attr.s(slots=True)
class HAMessageType27(HAPayload):
    """
    Long Range AIS Broadcast message
    """
    msg_type = bit_field(6, int, default=27, signed=False)
    repeat = bit_field(2, int, default=0, signed=False)
    mmsi = bit_field(30, int, from_converter=from_mmsi)

    accuracy = bit_field(1, bool, default=0, signed=False)
    raim = bit_field(1, bool, default=0, signed=False)
    status = bit_field(4, int, default=0, from_converter=NavigationStatus, to_converter=NavigationStatus, signed=False)
    lon = bit_field(18, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)
    lat = bit_field(17, float, from_converter=from_lat_lon_600, to_converter=to_lat_lon_600, default=0, signed=True)
    speed = bit_field(6, float, default=0, signed=False)
    course = bit_field(9, float, default=0, signed=False)
    gnss = bit_field(1, bool, default=0, signed=False)
    spare_1 = bit_field(1, bytes, default=b'', is_spare=True)

pyais.messages.MessageType27 = HAMessageType27
pyais.messages.MSG_CLASS[27] = pyais.messages.MessageType27


# Improve TagBlock to support any tag block

class HATagBlock(TagBlock):

    @classmethod
    def add_field(cls, field_name, field_code):
        """ Add a tag block field in addition to those existing by default.
        @:param field_name: Name of the tag block
        @:param field_code: Code of the tag block
        Note: this method must be called on class before instantiating it.
        """
        if field_name in cls.FIELD_CODES:
            raise ValueError(f"Field of name '{field_name}' is already defined")
        if field_code in cls.FIELD_NAMES:
            raise ValueError(f"Field of code '{field_code}' is already defined")

        cls.FIELD_CODES[field_name] = field_code
        cls.FIELD_NAMES[field_code] = field_name

    def __init__(self, raw: bytes) -> None:
        super().__init__(raw)

        # Initialize tag blocks added with add_field method
        for field_name in self.FIELD_CODES:
            if not hasattr(self, f"_{field_name}"):
                setattr(self, f"_{field_name}", None)

    def __repr__(self) -> str:
        if not self.initialized:
            return "TagBlock<uninitialized>"
        elif self.raw is None:
            return "TagBlock<None>"
        return f"TagBlock<{self.raw.decode()}>"

    def init(self) -> None:
        if self.raw is None:
            self.initialized = True
        else:
            super().init()

    def set_field_value(self, field_name, value):
        if field_name not in self.FIELD_CODES:
            raise ValueError(f"Field of name '{field_name}' is not defined")
        setattr(self, f"_{field_name}", value)

    @error_if_uninitialized
    def asdict(self) -> typing.Dict[str, typing.Any]:
        """Return a dictionary representation of the TagBlock."""
        res = {'raw': self.raw}
        for field_name in self.FIELD_CODES:
            res[field_name] = getattr(self, f"_{field_name}")
        return res

    def to_raw(self) -> bytes:
        """Convert a tag block to raw bytes"""
        fields = []
        if self._group is not None:
            fields.append(f"g:{self._group}")
        for field_name, field_code in self.FIELD_CODES.items():
            value = getattr(self, f"_{field_name}")
            if value is not None:
                fields.append(f"{field_code}:{value}")

        if not fields:
            raise ValueError('can not convert empty tag block to bytes (forgot to call .init()?)')

        payload_str = ','.join(fields)
        payload = payload_str.encode()

        chk_int = checksum(payload)
        chk = f"{chk_int:02X}".encode()

        return payload + ASTERISK + chk

pyais.messages.TagBlock = HATagBlock
pyais.TagBlock = HATagBlock






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.context.session_context import SessionContext
import logging
from holado_report.report.reports.base_report import BaseReport
from holado.holado_config import Config
from holado_system.system.filesystem.file import File
# from holado_core.scenario.scenario_duration_manager import ScenarioDurationManager

logger = logging.getLogger(__name__)




class ReportManager(BaseReport):
    """ Manage reports of current session
    """
    TFeatureReport = None
    StepTools = None
    ScenarioTools = None
    
    
    @classmethod
    def _get_test_server_client(cls):
        return SessionContext.instance().test_server_client
    
    
    def __init__(self):
        super().__init__()
        
        self.__multitask_manager = None
        
        # Auto configuration
        self.configure()
    
    def configure(self):
        from holado_report.report.reports.feature_report import FeatureReport
        ReportManager.TFeatureReport = FeatureReport
        
        from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
        ReportManager.StepTools = BehaveStepTools
        
        from holado_test.behave.scenario.behave_scenario_tools import BehaveScenarioTools
        ReportManager.ScenarioTools = BehaveScenarioTools
    
    def initialize(self, multitask_manager):
        self.__multitask_manager = multitask_manager
        
    def initialize_reports(self):
        from holado_report.report.builders.detailed_scenario_report_builder import DetailedScenarioReportBuilder
        from holado_report.report.builders.summary_report_builder import SummaryReportBuilder
        from holado_report.report.builders.summary_scenario_report_builder import SummaryScenarioReportBuilder
        from holado_report.report.builders.short_scenario_report_builder import ShortScenarioReportBuilder
        from holado_report.report.builders.failure_report_builder import FailureReportBuilder
        from holado_report.report.builders.summary_scenario_by_category_report_builder import SummaryScenarioByCategoryReportBuilder
        from holado_report.report.builders.summary_by_category_report_builder import SummaryByCategoryReportBuilder
        
        # self.set_execution_historic()
        
        if self.has_report_path:
            # fn = self.get_path("execution_historic.json")
            # self.add_report_builder(ExecutionHistoricReportBuilder(self.execution_historic, fn))
            
            fn = self.get_path("report_summary_scenario_failed.txt")
            self.add_report_builder(SummaryScenarioReportBuilder(fn, exclude_statuses=['Passed']))
            
            fn = self.get_path("report_summary_scenario_failed_by_category.txt")
            self.add_report_builder(SummaryScenarioByCategoryReportBuilder(fn, exclude_categories=['Always Success']))
            
            fn = self.get_path("report_summary_scenario_all.txt")
            self.add_report_builder(SummaryScenarioReportBuilder(fn))
            
            fn = self.get_path("report_failures.xml")
            self.add_report_builder(FailureReportBuilder(fn))
            
            fn = self.get_path("report_short_scenario_failed.txt")
            self.add_report_builder(ShortScenarioReportBuilder(fn, exclude_statuses=['Passed']))
            
            fn = self.get_path(Config.campaign_manager_import_report_name)  # @UndefinedVariable
            self.add_report_builder(DetailedScenarioReportBuilder(fn))
            
            fn = self.get_path("report_summary.txt")
            self.add_report_builder(SummaryReportBuilder(fn))
        
            fn = self.get_path("report_summary_by_category.txt")
            self.add_report_builder(SummaryByCategoryReportBuilder(fn))
        
    @property
    def __feature_reports(self):
        return self.children_reports("feature")
    
    @property
    def current_feature_report(self):
        # if self.__feature_reports:
        #     return self.__feature_reports[-1][1]
        # else:
        #     return None
        if SessionContext.instance().has_feature_context():
            feature_context = SessionContext.instance().get_feature_context()
            if feature_context.has_object("feature_report"):
                return feature_context.get_object("feature_report")
        return None
    
    @property
    def current_scenario_report(self):
        cfr = self.current_feature_report
        if cfr:
            return cfr.current_scenario_report
        else:
            return None
    
    def new_session(self, report_path):
        self.report_path = report_path
        if self.has_report_path:
            SessionContext.instance().path_manager.makedirs(self.report_path, is_directory=True)
        
        self.initialize_reports()
    
    def before_all(self):
        super().before_all()
    
    def before_feature(self, feature_context, feature):
        # Create new feature report
        fr = ReportManager.TFeatureReport(self, feature_context, feature)
        self.add_child_report(fr, "feature", feature.name, feature.filename)
        fr.initialize_reports()
        
        # Process self reports
        super().before_feature(feature_context, feature, feature_report=fr)
        
        # Store feature report in current feature context
        self.__multitask_manager.get_feature_context().set_object("feature_report", fr)
        
    def before_scenario(self, scenario_context, scenario):
        # Create new scenario report in feature report
        sr = self.current_feature_report.new_scenario_report(scenario_context, scenario)
        
        # Process feature report
        self.current_feature_report.before_scenario(scenario_context, scenario, scenario_report=sr)
        
        # Process self reports
        super().before_scenario(scenario_context, scenario, scenario_report=sr)
        
        # Change active log file to scenario one
        self.__enter_current_scenario_log_file()
    
    def before_step(self, step_context, step, step_level):
        super().before_step(step_context, step, step_level)
        
        # Note: A step can be executed without scenario report (ex: post processes after scenario) 
        if self.current_scenario_report:
            self.current_scenario_report.before_step(step_context, step, step_level)
    
    def after_step(self, step_context, step, step_level):
        super().after_step(step_context, step, step_level)
        
        # Note: A step can be executed without scenario report (ex: post processes after scenario) 
        if self.current_scenario_report:
            self.current_scenario_report.after_step(step_context, step, step_level)
    
    def after_scenario(self, scenario):
        self.current_feature_report.after_scenario(scenario, self.current_scenario_report)
        super().after_scenario(scenario, self.current_scenario_report)
        
        # Change active log file to root one
        self.__leave_current_scenario_log_file()
        
        self.current_scenario_report.release_resources()
    
    def after_feature(self, feature):
        super().after_feature(feature, self.current_feature_report)
        
        self.current_feature_report.release_resources()
    
    def after_all(self):
        super().after_all(build_reports=True)
            
        # Create files using execution historic as input
        # fn_eh = self.get_path("execution_historic.json")
        # sdm = ScenarioDurationManager()
        # sdm.import_execution_historic(fn_eh)
        #
        # fn = self.get_path("scenario_durations.csv")
        # scenario_duration_limits = sdm.compute_scenario_duration_limits()
        # sdm.create_file_scenario_duration_limits(fn, scenario_duration_limits)
        #
        # fn = self.get_path("scenario_duration_tags.csv")
        # duration_limit_tags = [(1, "fast"), (5, "rapid"), (60, "slow")]
        # scenario_duration_tags = sdm.compute_scenario_duration_tags(duration_limit_tags, "long", missing_tag=True, new_tag=True, unchanged_tag=True, with_failed=True)
        # sdm.create_file_scenario_duration_tags(fn, scenario_duration_tags)
        
        # Update campaigns stored in test server
        if self._get_test_server_client().is_available:
            self._get_test_server_client().update_stored_campaigns()
        
    def __enter_current_scenario_log_file(self):
        if SessionContext.instance().log_manager.in_file and self.has_report_path:
            log_filename = self.current_scenario_report.get_path("logs", "report.log")
            SessionContext.instance().path_manager.makedirs(log_filename)
            # Note: do_remove_other_file_handlers is set to True, in case leaving previous scenario log file has failed.
            #       Normally this case shouldn't appear but it was already observed. And when it appeared, execution slowdowned drastically.
            SessionContext.instance().log_manager.enter_log_file(log_filename, do_remove_other_file_handlers=True)
        
    def __leave_current_scenario_log_file(self):
        if SessionContext.instance().log_manager.in_file and self.has_report_path:
            log_filename = self.current_scenario_report.get_path("logs", "report.log")
            SessionContext.instance().log_manager.leave_log_file(log_filename, do_remove_log_file=True)
    
    def report_abort(self, reason):
        if self.has_report_path:
            log_filename = self.get_path("logs", "abort.log")
            SessionContext.instance().path_manager.makedirs(log_filename)
            File.write_file_with_content(log_filename, reason)
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado.common.handlers.object import DeleteableObject
import io
import paramiko
from holado_core.common.tools.path_manager import PathManager
from holado.common.context.session_context import SessionContext

logger = logging.getLogger(__name__)

try:
    import pysftp
    with_pysftp = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"SFTPClient is not available. Initialization failed on error: {exc}")
    with_pysftp = False


class SFTPClient(DeleteableObject):
    @classmethod
    def is_available(cls):
        return with_pysftp
    
    def __init__(self, **connection_kwargs):
        super().__init__("sFTP client")
        
        self.__connection = None
        if len(connection_kwargs) > 0:
            self.connect(**connection_kwargs)
    
    def _delete_object(self):
        if self.__connection:
            self.close_connection()
            
    @property
    def connection(self) -> pysftp.Connection:
        return self.__connection
    
    def connect(self, **connection_kwargs):
        kwargs = dict(connection_kwargs)
        
        if Tools.has_sub_kwargs(kwargs, "cnopts."):
            cnopts_kwargs = Tools.pop_sub_kwargs(kwargs, "cnopts.")
            cnopts = pysftp.CnOpts()
            if "hostkeys" in cnopts_kwargs:
                cnopts.hostkeys = cnopts_kwargs.pop("hostkeys")
            if "log" in cnopts_kwargs:
                cnopts.log = cnopts_kwargs.pop("log")
                SessionContext.instance().path_manager.makedirs(cnopts.log)
            if cnopts_kwargs:
                raise FunctionalException(f"Unmanaged cnopts parameters: {cnopts_kwargs}")
            kwargs["cnopts"] = cnopts
        
        if 'private_key' in kwargs:
            pk = kwargs['private_key']
            if isinstance(pk, str) and pk.startswith('-----'):
                # Replace private key text by a paramiko RSAKey
                privkey = io.StringIO(pk)
                key = paramiko.RSAKey.from_private_key(privkey)
                kwargs['private_key'] = key
        
        local_path = kwargs.pop("local_path") if "local_path" in kwargs else None
        if local_path is not None:
            SessionContext.instance().path_manager.makedirs(local_path)
        
        self.__connection = pysftp.Connection(**kwargs)
        
        if local_path:
            self.lcd(local_path)
        
    def close_connection(self):
        if self.__connection is None:
            raise TechnicalException("Client is not connected")
        
        try:
            self.__connection.close()
        except Exception as exc:  # @UnusedVariable
            #TODO: When this warning is logged during self.__del__, the log is cleared before, thus it is commented
            # logger.warn(f"Error catched while closing RabbitMQ client connection:\n{Tools.represent_exception(exc)}")
            pass
        finally:
            self.__connection = None
        
    @classmethod
    def lcd(cls, path):
        pysftp.cd(path)
    


from django.contrib.auth.models import User
from django_grpc_framework import generics
from api_grpc.api1.serializers import UserProtoSerializer


class UserService(generics.ModelService):
    """
    gRPC service that allows users to be retrieved or updated.
    """
    queryset = User.objects.all().order_by('-date_joined')
    serializer_class = UserProtoSerializer



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.element_exception import ElementNotFoundException
from holado_core.common.exceptions.technical_exception import TechnicalException
import logging
from holado.common.handlers.undefined import undefined_argument
from holado_core.common.handlers.features.resource_by_name import FeatureClassResourceByName,\
    FeatureObjectResourceByName
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


class FeatureClassResourceByType(FeatureClassResourceByName):
    @classmethod
    def _get_class_resource_description(cls, plural=False):
        if plural:
            return 'class resources by type'
        else:
            return 'class resource by type'
    
    @classmethod
    def register_resource_for_type_in_class(cls, name, is_class_func, is_obj_func, resource, index=-1):
        super().register_resource_in_class(name, (is_class_func, is_obj_func, resource), index=index)
    
    @classmethod
    def import_registered_resources_in_class(cls, class_=None, obj=None):
        if (class_ is not None and issubclass(class_, FeatureClassResourceByType)
            or obj is not None 
                and (isinstance(obj, FeatureClassResourceByType) or isinstance(obj, FeatureObjectResourceByType) ) ):
            super(FeatureClassResourceByType, cls).import_registered_resources_in_class(class_=class_, obj=obj)
        else:
            raise TechnicalException(f"Unmanaged import registered {cls._get_class_resource_description(True)} from (class_={class_} ; type(obj)={Typing.get_object_class_fullname(obj)})")
    
    @classmethod
    def get_class_resource_for_type(cls, class_=None, obj=undefined_argument, raise_if_not_found=True):
        # Replace a 'typing' class description by a real class
        #TODO: manage Union class_ by looping on union types
        real_class = Typing.get_origin_type(class_)
        
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"[{Typing.get_class_fullname(cls)}] Finding class resource in {cls.get_class_resource_names()} for type '{Typing.get_class_fullname(class_)}'" + (f" and object of type '{Typing.get_object_class_fullname(obj)}'" if obj is not undefined_argument else ""))
        res = None
        for resource_name in cls.get_class_resource_names():
            is_class_func, is_obj_func, resource = cls.get_class_resource(resource_name)
            try:
                # Note: start by using is_obj_func and/or obj as it can be more accurate
                if is_obj_func is not None and obj is not undefined_argument:
                    if is_obj_func(obj):
                        res = resource
                elif is_class_func is not None:
                    if obj is not undefined_argument:
                        if is_class_func(type(obj)):
                            res = resource
                    elif real_class is not None:
                        if is_class_func(real_class):
                            res = resource
            except Exception:
                real_class_descr = '' if real_class == class_ else f' ; real class={real_class}'
                # raise TechnicalException(f"Error while getting {cls._get_class_resource_description()} named '{resource_name}' and associated to (class={class_}{real_class_descr} ; type(obj)={Typing.get_object_class_fullname(obj)}) in registered {cls._get_class_resource_description(True)} {list(cls._get_class_registered_resources())}") from exc
                if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"Error while getting {cls._get_class_resource_description()} named '{resource_name}' and associated to (class={class_}{real_class_descr} ; type(obj)={Typing.get_object_class_fullname(obj)}) in registered {cls._get_class_resource_description(True)} {list(cls._get_class_registered_resources())}")
                    
            if res is not None:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"[{Typing.get_class_fullname(cls)}] Found class resource '{resource_name}' for type '{Typing.get_class_fullname(class_)}'" + (f" and object of type '{Typing.get_object_class_fullname(obj)}'" if obj is not undefined_argument else ""))
                return res
            else:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"[{Typing.get_class_fullname(cls)}] Class resource '{resource_name}' doesn't match for type '{Typing.get_class_fullname(class_)}'" + (f" and object of type '{Typing.get_object_class_fullname(obj)}'" if obj is not undefined_argument else ""))
       
        if raise_if_not_found:
            raise ElementNotFoundException(f"Not found {cls._get_class_resource_description()} associated to (class={class_} ; type(obj)={Typing.get_object_class_fullname(obj)}) in registered {cls._get_class_resource_description(True)} {list(cls._get_class_registered_resources())}")
        else:
            return None


class FeatureObjectResourceByType(FeatureObjectResourceByName):
    def __init__(self):
        super().__init__()
        
    def _get_resource_description(self, plural=False):
        if plural:
            return 'resources by type'
        else:
            return 'resource by type'
        
    def register_resource_for_type(self, name, is_class_func, is_obj_func, resource, index=-1):
        self.register_resource(name, (is_class_func, is_obj_func, resource))
        
    def import_registered_resources(self, class_=None, obj=None):
        if (class_ is not None and issubclass(class_, FeatureClassResourceByType)
            or obj is not None 
                and (isinstance(obj, FeatureClassResourceByType) or isinstance(obj, FeatureObjectResourceByType) ) ):
            super().import_registered_resources(class_=class_, obj=obj)
        else:
            raise TechnicalException(f"Unmanaged import registered {self._get_resource_description(True)} from (class_={class_} ; type(obj)={Typing.get_object_class_fullname(obj)})")
        
    def get_resource_for_type(self, class_=None, obj=undefined_argument, raise_if_not_found=True):
        # Replace a 'typing' class description by a real class
        #TODO: manage Union class_ by looping on union types
        real_class = Typing.get_origin_type(class_)
        
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"[{Typing.get_object_class_fullname(self)}] Finding resource in {self.get_resource_names()} matching for type '{Typing.get_class_fullname(class_)}'" + (f" and object of type '{Typing.get_object_class_fullname(obj)}'" if obj is not undefined_argument else ""))
        res = None
        for resource_name in self.get_resource_names():
            is_class_func, is_obj_func, resource = self.get_resource(resource_name)
            try:
                # Note: start by using is_obj_func and/or obj as it can be more accurate
                if is_obj_func is not None and obj is not undefined_argument:
                    if is_obj_func(obj):
                        res = resource
                    else:
                        # Current resource is not matchine
                        pass
                elif is_class_func is not None:
                    if obj is not undefined_argument:
                        if is_class_func(type(obj)):
                            res = resource
                        else:
                            # Current resource is not matchine
                            pass
                    elif real_class is not None:
                        if is_class_func(real_class):
                            res = resource
                    else:
                        raise TechnicalException(f"Unexpected case")
                else:
                    raise TechnicalException(f"Unexpected case")
            except Exception as exc:
                raise TechnicalException(f"Error while getting {self._get_resource_description()} named '{resource_name}' and associated to (class={class_} ; type(obj)={Typing.get_object_class_fullname(obj)}) in registered {self._get_resource_description(True)} {list(self._registered_resources.keys())}") from exc
            
            if res is not None:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"[{Typing.get_object_class_fullname(self)}] Found resource '{resource_name}' for type '{Typing.get_class_fullname(class_)}'" + (f" and object of type '{Typing.get_object_class_fullname(obj)}'" if obj is not undefined_argument else ""))
                return res
            else:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"[{Typing.get_object_class_fullname(self)}] Resource '{resource_name}' doesn't match for type '{Typing.get_class_fullname(class_)}'" + (f" and object of type '{Typing.get_object_class_fullname(obj)}'" if obj is not undefined_argument else ""))
        
        if raise_if_not_found:
            raise ElementNotFoundException(f"Not found {self._get_resource_description()} associated to (class={class_} ; type(obj)={Typing.get_object_class_fullname(obj)}) in registered {self._get_resource_description(True)} {list(self._registered_resources.keys())}")
        else:
            return None

class FeatureObjectOverClassResourceByType(FeatureObjectResourceByType):
    def __init__(self):
        super().__init__()

    def get_resource_for_type(self, class_=None, obj=undefined_argument, raise_if_not_found=True):
        res = super().get_resource_for_type(class_, obj, raise_if_not_found=False)
        if not res:
            res = self.get_class_resource_for_type(class_, obj, raise_if_not_found=False)
        
        if not res and raise_if_not_found:
            raise ElementNotFoundException(f"Not found {self._get_resource_description()} associated to (class={class_} ; type(obj)={Typing.get_object_class_fullname(obj)}) in registered {self._get_resource_description(True)} {list(self._registered_resources.keys())} and registered {self._get_class_resource_description(True)} {list(self._get_class_registered_resources().keys())}")
        return res

class FeatureObjectAndClassResourceByType(FeatureObjectOverClassResourceByType, FeatureClassResourceByType):
    def __init__(self):
        FeatureObjectOverClassResourceByType.__init__(self)





# -*- coding: utf-8 -*-

import logging
from holado_test.behave.behave import *  # @UnusedWildImport
from holado.common.context.session_context import SessionContext
from config.config_manager import TSConfigManager  # @UnresolvedImport


logger = logging.getLogger(__name__)


def __get_config_manager() -> TSConfigManager:
    return SessionContext.instance().config_manager


@Given(r"ensure system is configured with default settings")
def step_impl(context):
    __get_config_manager().configure_system_with_default_settings()







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import object
import logging
from holado_value.common.tools.value_types import ValueTypes
from holado_core.common.exceptions.technical_exception import TechnicalException
import re
import importlib
from holado_core.common.tools.converters.converter import Converter
import types
from holado_core.common.tools.tools import Tools
from holado_multitask.multithreading.reflection import traceback
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
from holado.holado_config import Config
from holado_multitask.multitasking.multitask_manager import MultitaskManager
from holado_python.standard_library.typing import Typing
from holado.common.handlers.undefined import undefined_value, not_applicable

logger = logging.getLogger(__name__)


class ExpressionEvaluator(object):

    def __init__(self):
        self.__dynamic_text_manager = None
        self.__text_interpreter = None
        self.__unique_value_manager = None
        self.__variable_manager = None
        
        # self.__modules_already_imported = []
        
    def initialize(self, dynamic_text_manager, unique_value_manager, text_interpreter, variable_manager):
        self.__dynamic_text_manager = dynamic_text_manager
        self.__text_interpreter = text_interpreter
        self.__unique_value_manager = unique_value_manager
        self.__variable_manager = variable_manager
            
    def evaluate_expression(self, expression, unescape_string_method=None, eval_params=EvaluateParameters.default(), log_level=logging.DEBUG):
        value_type, value, value_to_eval = self.extract_expression_information(expression, unescape_string_method, eval_params=eval_params)
        if value_to_eval is not undefined_value:
            if ValueTypes.is_string(value_type):
                eval_params = eval_params.with_result_type(str)
            res_type, res = self.evaluate_expression_of_information(value_type, value_to_eval, eval_params=eval_params, log_level=log_level)
            if Tools.do_log(logger, log_level):
                logger.log(log_level, f"Evaluate expression [{expression}] => ({value_type.name}, {value}, '{value_to_eval}') => ({res_type.name}) [{res}] (type: {Typing.get_object_class_fullname(res)})    (with evaluate parameters: {eval_params})")
        else:
            res_type, res = value_type, value
            if Tools.do_log(logger, log_level):
                logger.log(log_level, f"Evaluate expression [{expression}] => ({res_type.name}) [{res}] (type: {Typing.get_object_class_fullname(res)})    (with evaluate parameters: {eval_params})")
        return res_type, res

    def evaluate_python_expression(self, expression, eval_params=EvaluateParameters.default(), log_level=logging.DEBUG):
        res_type, res = self.evaluate_expression_of_information(ValueTypes.Symbol, expression, eval_params=eval_params, log_level=log_level)
        if Tools.do_log(logger, log_level):
            logger.log(log_level, f"Evaluate python expression [{expression}] => ({res_type.name}) [{res}] (type: {Typing.get_object_class_fullname(res)})    (with evaluate parameters: {eval_params})")
        return res_type, res

    def extract_expression_information(self, expression, unescape_string_method=None, eval_params=EvaluateParameters.default(), log_level=logging.TRACE):  # @UndefinedVariable
        res = self.__extract_expression_information(expression, unescape_string_method=unescape_string_method, eval_params=eval_params)
        if Tools.do_log(logger, log_level):
            logger.log(log_level, f"Extract expression information [{expression}] => {res}")
        return res

    def __extract_expression_information(self, expression, unescape_string_method=None, eval_params=EvaluateParameters.default()):
        if isinstance(expression, str):
            return self.__extract_expression_information_str(expression, unescape_string_method=unescape_string_method, eval_params=eval_params)
        else:
            return (*self.__extract_value_information(expression, from_symbol=False, from_str=False), undefined_value)

    def __extract_expression_information_str(self, expression, unescape_string_method=None, eval_params=EvaluateParameters.default()):
        if unescape_string_method is None:
            from holado_test.scenario.step_tools import StepTools
            unescape_string_method = StepTools.unescape_string
            
        try:
            expr_content = expression.strip()
            
            # Manage usual types
            value_type, value = self.__extract_value_information(expr_content, from_symbol=not eval_params.result_is_str, from_str=not eval_params.result_is_str)
            if value_type != ValueTypes.String:
                return (value_type, value, undefined_value)
            
            # Manage strings
            for end_sym, val_type in [(Config.DYNAMIC_SYMBOL, ValueTypes.DynamicString), 
                                      (Config.THREAD_DYNAMIC_SYMBOL, ValueTypes.ThreadDynamicString), 
                                      (Config.UNIQUE_SYMBOL, ValueTypes.UniqueString), 
                                      ("", ValueTypes.String)]:
                if expr_content.startswith("'") and expr_content.endswith("'" + end_sym):
                    text = expr_content[1:-1-len(end_sym)]
                    if unescape_string_method is not None:
                        text = unescape_string_method(text)
                    return (val_type, undefined_value, text)
            
            # Else consider content as a symbol
            return (ValueTypes.Symbol, undefined_value, expr_content)
        except Exception as exc:
            raise TechnicalException(f"Error while extracting expression information in expression [{expression}] (type: {Typing.get_object_class_fullname(expression)})") from exc

    def extract_value_information(self, value, *, from_symbol=True, from_str=True):
        return self.__extract_value_information(value, from_symbol=from_symbol, from_str=from_str)

    def __extract_value_information(self, value, *, from_symbol=False, from_str=False):
        if value is undefined_value:
            return (ValueTypes.Undefined, undefined_value)
        elif value is not_applicable:
            return (ValueTypes.NotApplicable, not_applicable)
        elif value is None:
            return (ValueTypes.Null, None)
        elif isinstance(value, bool):
            return (ValueTypes.Boolean, value)
        elif isinstance(value, int):
            return (ValueTypes.Integer, value)
        elif isinstance(value, float):
            return (ValueTypes.Float, value)
        elif isinstance(value, str):
            if from_symbol:
                if value == Config.NONE_SYMBOL:
                    return (ValueTypes.Null, None)
                elif value == Config.NOT_APPLICABLE_SYMBOL:
                    return (ValueTypes.NotApplicable, not_applicable)
            
            if from_str:
                if Converter.is_boolean(value):
                    return (ValueTypes.Boolean, Converter.to_boolean(value))
                elif Converter.is_integer(value):
                    return (ValueTypes.Integer, int(value))
                elif Converter.is_float(value):
                    return (ValueTypes.Float, float(value))
            
            return (ValueTypes.String, value)
        else:
            return (ValueTypes.Generic, value)
        
    def evaluate_expression_of_information(self, value_type, value, eval_params=EvaluateParameters.default(), log_level=logging.DEBUG):
        if Tools.do_log(logger, logging.TRACE, max_log_level=log_level):  # @UndefinedVariable
            logger.log(Tools.do_log_level(logging.TRACE, log_level), f"Evaluating expression of information ({value_type.name}, {value}, {eval_params})")  # @UndefinedVariable
        res_type, res = self.__evaluate_expression_of_information(value_type, value, eval_params=eval_params, log_level=log_level)
        if Tools.do_log(logger, logging.TRACE, max_log_level=log_level):  # @UndefinedVariable
            logger.log(Tools.do_log_level(logging.TRACE, log_level), f"Evaluate expression of information ({value_type.name}, {value}) => ({res_type.name}) [{res}] (type: {Typing.get_object_class_fullname(res)})    (with evaluate parameters: {eval_params})")  # @UndefinedVariable
        return res_type, res
        
    def __evaluate_expression_of_information(self, value_type, value, eval_params=EvaluateParameters.default(), log_level=logging.DEBUG):
        if value_type in [ValueTypes.NotApplicable]:
            return (value_type, not_applicable)
        elif value_type in [ValueTypes.Null]:
            return (value_type, None)
        elif value_type in [ValueTypes.Boolean, ValueTypes.Integer, ValueTypes.Float, ValueTypes.Generic]:
            return (value_type, value)
        elif ValueTypes.is_string(value_type):
            eval_params = eval_params.with_result_type(str)
            return self.__evaluate_expression_of_information_string(value_type, value, eval_params=eval_params, log_level=log_level)
        elif value_type in [ValueTypes.Symbol]:
            return self.__evaluate_expression_of_information_symbol(value, eval_params=eval_params, log_level=log_level)
        else:
            return self.__extract_value_information(value)
        
    def __evaluate_expression_of_information_string(self, value_type, value, eval_params=EvaluateParameters.default_without_raise(), log_level=logging.DEBUG):
        if eval_params.do_interpret:
            res = self.__text_interpreter.interpret(value, eval_params=eval_params, log_level=log_level)
        else:
            res = value
            
        if isinstance(res, bytes):
            res = res.decode('utf-8')
        elif not isinstance(res, str):
            res = str(res)
            
        if value_type == ValueTypes.DynamicString:
            res = self.__dynamic_text_manager.get(res)
        elif value_type == ValueTypes.ThreadDynamicString:
            res = self.__dynamic_text_manager.get(res, scope=MultitaskManager.get_thread_id())
        elif value_type == ValueTypes.UniqueString:
            res = res + self.__unique_value_manager.new_string(padding_length=Config.unique_string_padding_length)
        
        return self.extract_value_information(res, from_symbol=not eval_params.result_is_str, from_str=not eval_params.result_is_str)
        
    def __evaluate_expression_of_information_symbol(self, value, eval_params=EvaluateParameters.default(), log_level=logging.DEBUG):
        if Tools.do_log(logger, logging.TRACE, max_log_level=log_level):  # @UndefinedVariable
            logger.log(Tools.do_log_level(logging.TRACE, log_level), f"Evaluating symbol ([{value}], {eval_params})")  # @UndefinedVariable
        res = value
        
        # Replace explicit interpret parts
        if eval_params.do_interpret and isinstance(res, str):
            res, _, is_evaluated = self.__evaluate_interpret_expression(res, eval_params=eval_params, log_level=log_level)
            if is_evaluated:
                # If symbol is already evaluated, stop evaluation process
                # Note: evaluation of string was already done if asked, thus just evaluate symbols
                res_type, res = self.__extract_value_information(res, from_symbol=not eval_params.result_is_str, from_str=not eval_params.result_is_str)
                if Tools.do_log(logger, logging.TRACE, max_log_level=log_level):  # @UndefinedVariable
                    logger.log(Tools.do_log_level(logging.TRACE, log_level), f"Evaluate symbol [{value}]  =>  ({res_type.name}) [{res}] (type: {Typing.get_object_class_fullname(res)})    (after interpret, with evaluate parameters: {eval_params})")  # @UndefinedVariable
                return res_type, res
            # if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            #     logger.trace(f"Evaluate symbol [{value}] - after interpret:  [{res}] (type: {Typing.get_object_class_fullname(res)})    (with evaluate parameters: {eval_params})")
        
        # Evaluate variable expression
        # Note: if an eval must be done after interpret, variable shouldn't be evaluated unless it can makes failing eval
        if eval_params.do_eval_variable and isinstance(res, str) and len(res) > 0:
            result, is_evaluated = self.__evaluate_variable_expression(res, eval_params=eval_params, log_level=log_level)
            if is_evaluated:
                # If symbol is a variable expression, don't evaluate the result of the variable
                res_type, res = self.__extract_value_information(result, from_symbol=False, from_str=False)
                if Tools.do_log(logger, logging.TRACE, max_log_level=log_level):  # @UndefinedVariable
                    logger.log(Tools.do_log_level(logging.TRACE, log_level), f"Evaluate symbol [{value}]  =>  ({res_type.name}) [{res}] (type: {Typing.get_object_class_fullname(res)})    (after variable evaluation, with evaluate parameters: {eval_params})")  # @UndefinedVariable
                return res_type, res
        
        # Evaluate expression
        # if eval_params.do_eval and (isinstance(res, str) or isinstance(res, bytes)):    # Commented as it doesn't work currently with bytes
        if eval_params.do_eval and isinstance(res, str):
            res = self.__evaluate_expression(res, locals_={}, eval_params=eval_params, log_level=log_level)
            
        # logger.debug(f"Evaluate symbol [{value}] => [{res}] (type: {Typing.get_object_class_fullname(res)})")
        # Note: evaluation of string was already done if asked, thus just evaluate symbols
        res_type, res = self.__extract_value_information(res, from_symbol=not eval_params.result_is_str, from_str=not eval_params.result_is_str)
        if Tools.do_log(logger, logging.TRACE, max_log_level=log_level):  # @UndefinedVariable
            logger.log(Tools.do_log_level(logging.TRACE, log_level), f"Evaluate symbol [{value}]  =>  [{res}] (type: {Typing.get_object_class_fullname(res)})    (after full evaluation, with evaluate parameters: {eval_params})")  # @UndefinedVariable
        return res_type, res
        
    def __evaluate_interpret_expression(self, value, eval_params=EvaluateParameters.default(), log_level=logging.DEBUG):
        result, is_interpreted, is_evaluated = value, False, False
        
        nb_sections = self.__text_interpreter.get_number_of_sections(value)
        if nb_sections > 0:
            # If eval is authorized, and many sections exists, prefer interpret sections and eval result
            if eval_params.do_eval and nb_sections > 1:
                result, locals_ = self.__text_interpreter._interpret_sections(value, eval_params=eval_params, log_level=log_level)
                if result != value:
                    result = self.__evaluate_expression(result, locals_, eval_params=eval_params, log_level=log_level)
                    is_evaluated = True
            else:
                result = self.__text_interpreter.interpret(value, eval_params=eval_params, log_level=log_level)
            is_interpreted = result != value
        
        # Manage recursive interpret, only in case of first interpret and if it wasn't evaluated
        if eval_params.do_interpret_recursively and is_interpreted and not is_evaluated and isinstance(result, str):
            new_result, new_is_interpreted, new_is_evaluated = self.__evaluate_interpret_expression(result, eval_params.with_raise_on_interpret_error(False), log_level=log_level)
            if new_is_interpreted:
                result = new_result
                is_evaluated = new_is_evaluated
        
        return result, is_interpreted, is_evaluated
        
    def __evaluate_variable_expression(self, value, eval_params=EvaluateParameters.default(), log_level=logging.DEBUG):
        is_evaluated, result = False, None
        
        if self.__variable_manager.exists_variable(value):
            result = self.__variable_manager.get_variable_value(value)
            is_evaluated = True
        elif self.__variable_manager.is_variable_expression(value):
            try:
                result = self.__variable_manager.eval_variable_expression(value)
            except Exception as exc:
                if eval_params.raise_on_variable_eval_error:
                    raise exc
                else:
                    if Tools.do_log(logger, log_level):
                        logger.log(log_level, f"Error while evaluating variable expression [{value}]: {Tools.represent_exception(exc)}\n -> traceback:\n{traceback.represent_stack(indent=4)}")
            else:
                is_evaluated = True
        
        # Manage recursive evaluation, only in case of first variable evaluation
        if eval_params.do_eval_variable_recursively and is_evaluated and isinstance(result, str) and result != value:
            new_result, new_is_evaluated = self.__evaluate_variable_expression(result, eval_params.with_raise_on_variable_eval_error(False), log_level=log_level)
            if new_is_evaluated:
                result = new_result
        
        return result, is_evaluated
        
    def __evaluate_expression(self, value, locals_, eval_params=EvaluateParameters.default(), log_level=logging.DEBUG):
        res = value
        while True:
            try:
                result = None
                if isinstance(value, str):
                    # Try first by replacing double semicolon
                    try:
                        txt_to_eval = value.replace("{{", "{").replace("}}", "}")
                        result = eval(txt_to_eval, globals(), locals_)
                    except SyntaxError as exc:
                        # Retry without replacing double semicolon
                        pass
                    
                if result is None:
                    result = eval(value, globals(), locals_)
            except NameError as exc:
                # logger.trace(f"Error while expression evaluation:\n    exception: {Tools.represent_exception(exc)}\n    globals: {globals()}\n    locals: {locals_}")
                # In case the name concerns a library, try to import the library
                m = re.match(r"name '(.*)' is not defined", str(exc))
                is_imported = False
                if m:
                    module_name = m.group(1)
                    is_imported = self.__import_module(module_name, locals_)
                    
                # If no import is done
                if not is_imported:
                    if Tools.do_log(logger, logging.DEBUG, max_log_level=log_level):
                        logger.log(Tools.do_log_level(logging.DEBUG, log_level), f"Error while evaluating expression [{value}]: {exc}")
                    break
            except Exception as exc:
                if eval_params.raise_on_eval_error:
                    raise exc
                else:
                    if Tools.do_log(logger, logging.TRACE, max_log_level=log_level):  # @UndefinedVariable
                        logger.log(Tools.do_log_level(logging.TRACE, log_level), f"Error while evaluating expression [{value}]: {Tools.represent_exception(exc)}")  # @UndefinedVariable
                    elif Tools.do_log(logger, logging.DEBUG, max_log_level=log_level):
                        if isinstance(exc, SyntaxError):
                            logger.log(Tools.do_log_level(logging.DEBUG, log_level), f"Error while evaluating expression [{value}]: {repr(exc)}")
                        else:
                            logger.log(Tools.do_log_level(logging.DEBUG, log_level), f"Error while evaluating expression [{value}]: [{Typing.get_object_class_fullname(exc)}] {exc}")
                break
            else:
                if not isinstance(result, types.ModuleType):
                    res = result
                break
            
        # logger.debug(f"Evaluate [{value}] => [{res}] (type: {Typing.get_object_class_fullname(res)})")
        return res
    
    # def __import_module(self, module_name, locals_):
    #     res = False
    #
    #     if module_name in locals_:
    #         raise TechnicalException(f"Module '{module_name}' is already imported (locals_: {locals_})")
    #
    #     if module_name not in self.__modules_already_imported:
    #         if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
    #             logger.trace(f"Importing module '{module_name}'")
    #         try:
    #             module = importlib.import_module(module_name)
    #             locals_[module_name] = module
    #         except Exception as exc2:
    #             if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
    #                 logger.trace(f"Failed to import module '{module_name}': {exc2}")
    #         else:
    #             if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
    #                 logger.trace(f"Imported module '{module_name}' for expression evaluation")
    #             res = True
    #         finally:
    #             self.__modules_already_imported.append(module_name)
    #     else:
    #         if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
    #             logger.trace(f"Skip import of module '{module_name}' since it was already imported")
    #
    #     return res
    def __import_module(self, module_name, locals_):
        res = False
        
        if module_name in locals_:
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"Skip import of module '{module_name}' since it was already imported")
        else:
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"Importing module '{module_name}'")
            try:
                module = importlib.import_module(module_name)
                locals_[module_name] = module
            except Exception as exc2:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"Failed to import module '{module_name}': {exc2}")
            else:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"Imported module '{module_name}' for expression evaluation")
                res = True
                
        return res
    
    def extract_string_value(self, expression, unescape_string_method=None, log_level=logging.DEBUG):
        value_type, value_str, value_to_eval = self.extract_expression_information(expression, unescape_string_method)
        if ValueTypes.is_string(value_type):
            if value_to_eval is not undefined_value:
                res = value_to_eval
            else:
                res = value_str
        else:
            res = expression
        if Tools.do_log(logger, log_level):
            logger.log(log_level, f"Extract string value [{expression}] => ({value_type}, '{value_str}') => [{res}] (type: {Typing.get_object_class_fullname(res)})")
        return res





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import threading
import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
import holado_multitask.multiprocessing as hmm
from holado.common.handlers.undefined import undefined_argument

logger = logging.getLogger(__name__)


class ProcessesManager(object):
    """
    Helper class to manage running processes.
    """
    
    
    def __init__(self, multitask_manager):
        self.__multitask_manager = multitask_manager
        self.__managed_processes_lock = threading.Lock()
        self.__managed_processes = {}
    
    def is_process_registered(self, name):
        with self.__managed_processes_lock:
            return name in self.__managed_processes
    
    def register_process(self, name, process):
        with self.__managed_processes_lock:
            if name in self.__managed_processes:
                logger.error(f"A process with name '{name}' is already registered, it will be overwritten: {self.__managed_processes[name]}")
            else:
                logger.debug(f"Registered process with name '{name}': {process}")
            self.__managed_processes[name] = process
    
    def unregister_process(self, name):
        with self.__managed_processes_lock:
            if name in self.__managed_processes:
                if self.__managed_processes[name].is_alive():
                    logger.warn(f"Unregistering process with name '{name}' whereas it is still alive")
                else:
                    logger.debug(f"Unregistered process with name '{name}'")
                del self.__managed_processes[name]
            else:
                raise Exception("Not registered process of name '{}'".format(name))


    def get_process(self, name):
        with self.__managed_processes_lock:
            if name in self.__managed_processes:
                return self.__managed_processes[name]
            else:
                raise Exception("Unregistered process of name '{}'".format(name))


    def interrupt_process(self, name, with_join=True, join_timeout=undefined_argument, raise_if_not_interruptable=True):
        process = self.get_process(name)
        if process.is_alive():
            if self.is_interruptable(process=process):
                try:
                    process.interrupt()
                except Exception as exc:
                    # logger.exception("Failed to interrupt process of name '{}'".format(name))
                    raise exc
                
                if with_join:
                    process.join(join_timeout)
            elif raise_if_not_interruptable:
                raise TechnicalException("Thread '{}' is not interruptable".format(name))
                


    def join_process(self, name, timeout=undefined_argument):
        process = self.get_process(name)
        if process.is_alive():
            process.join(timeout)


    def interrupt_all_processes(self, with_join=True, join_timeout=undefined_argument):
        """Interrupt all processes that are interruptable."""
        names = self.get_process_names()
        for name in names:
            try:
                if self.is_interruptable(name=name):
                    self.interrupt_process(name, with_join=False)
            except Exception:
                logger.exception("Failed to interrupt process of name '{}'".format(name))

        if with_join:
            self.join_all_processes(timeout=join_timeout, skip_not_interruptable=True)


    def join_all_processes(self, timeout=undefined_argument, skip_not_interruptable=False):
        names = self.get_process_names()
        for name in names:
            if skip_not_interruptable and self.is_interruptable(name=name):
                continue
            self.join_process(name, timeout=timeout)
    

    def unregister_all_processes(self, keep_alive=True):
        names = self.get_process_names()
        for name in names:
            process = self.get_process(name)
            if not keep_alive or keep_alive and not process.is_alive():
                self.unregister_process(name)
    

    def get_process_names(self):
        with self.__managed_processes_lock:
            return sorted(list( self.__managed_processes.keys() ))
    

    def get_alive_process_names(self):
        with self.__managed_processes_lock:
            res = sorted(list( self.__managed_processes.keys() ))
            res = [name for name in res if self.__managed_processes[name].is_alive()]
            

    def is_interruptable(self, process=None, name=None):
        if name is not None:
            process = self.get_process(name)
        if process is None:
            raise TechnicalException("At least one argument must be not None")    
        
        return isinstance(process, hmm.process.InterruptableProcess) and process.is_interruptable
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.tools.tools import Tools
import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


try:
    import pympler
    from pympler.tracker import SummaryTracker
    from pympler.tracker import ObjectTracker
    with_pympler = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"Pympler is not available for memory profiling. Initialization failed on error: {exc}")
    with_pympler = False


class MemoryProfiler():
    """
    Manage many methods to profile memory.
    It uses internally different libraries, depending on profiling method.
    """
    __stored_trackers = {}
    
    @classmethod
    def is_tracker_available(cls):
        return with_pympler
    
    @classmethod
    def create_or_reset_tracker_of_objects_summary_changes(cls, name):
        if name in cls.__stored_trackers:
            cls.reset_tracker(cls.__stored_trackers[name])
        else:
            cls.__stored_trackers[name] = cls.new_tracker_of_objects_summary_changes()
        return cls.__stored_trackers[name]
    
    @classmethod
    def create_or_reset_tracker_of_objects_changes(cls, name):
        if name in cls.__stored_trackers:
            cls.reset_tracker(cls.__stored_trackers[name])
        else:
            cls.__stored_trackers[name] = cls.new_tracker_of_objects_changes()
        return cls.__stored_trackers[name]
    
    @classmethod
    def new_tracker_of_objects_summary_changes(cls):
        if cls.is_tracker_available():
            return SummaryTracker()
        else:
            raise TechnicalException("Third library 'Pympler' is needed for this profiling method")
    
    @classmethod
    def new_tracker_of_objects_changes(cls):
        if cls.is_tracker_available():
            return ObjectTracker()
        else:
            raise TechnicalException("Third library 'Pympler' is needed for this profiling method")
    
    @classmethod
    def reset_tracker(cls, tracker):
        if isinstance(tracker, SummaryTracker):
            _ = tracker.diff()
        elif isinstance(tracker, ObjectTracker):
            _ = tracker.get_diff()
        else:
            raise TechnicalException(f"Unexpected tracker type '{Typing.get_object_class_fullname(tracker)}'")
    
    @classmethod
    def log_tracker_diff(cls, name=None, tracker=None, prefix=None, level=logging.DEBUG, logger_=None):
        if logger_ is None:
            logger_ = logger
        if not logger_.isEnabledFor(level):
            return
        
        if name is not None:
            tracker = cls.__stored_trackers.get(name)
        if tracker is None:
            raise TechnicalException(f"Parameter 'tracker' or 'name' must be defined")
        
        if isinstance(tracker, SummaryTracker):
            tracker_descr = "objects summary"
        elif isinstance(tracker, ObjectTracker):
            tracker_descr = "objects summary"
        else:
            raise TechnicalException(f"Unexpected tracker type '{Typing.get_object_class_fullname(tracker)}'")
        
        logger_.log(level, f"{prefix if prefix else ''}Preparing changes in {tracker_descr}...")
        formatted_diff = "\n".join(tracker.format_diff())
        logger_.log(level, f"{prefix if prefix else ''}Changes in {tracker_descr}:\n{formatted_diff}")
        





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import threading
import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
import holado_multitask.multithreading as hmm
from holado.common.handlers.undefined import default_context, undefined_argument
from holado_multitask.multitasking.multitask_manager import MultitaskManager

logger = logging.getLogger(__name__)


class ThreadsManager(object):
    """
    Manage threads run by HolAdo, so that each thread can use HolAdo services.
    """
    
    __has_native_id = hasattr(threading, '_HAVE_THREAD_NATIVE_ID') and threading._HAVE_THREAD_NATIVE_ID  # @UndefinedVariable
    
    @classmethod
    def has_native_id(cls):
        return cls.__has_native_id
    
    def __init__(self, multitask_manager):
        self.__multitask_manager = multitask_manager
        self.__managed_threads_lock = threading.Lock()
        self.__managed_threads_by_scope = {}
    
    def __get_scope_name(self, scope, create_if_doesnt_exist=True):
        res = MultitaskManager.get_scope_name(scope)
        
        if create_if_doesnt_exist:
            with self.__managed_threads_lock:
                if res not in self.__managed_threads_by_scope:
                    self.__managed_threads_by_scope[res] = {}
        
        return res
    
    def is_thread_registered(self, name, scope=default_context):
        scope_name = self.__get_scope_name(scope, create_if_doesnt_exist=False)
        with self.__managed_threads_lock:
            return scope_name in self.__managed_threads_by_scope and name in self.__managed_threads_by_scope[scope_name]
        
    def register_thread(self, name, thread, scope=default_context):
        scope_name = self.__get_scope_name(scope)
        with self.__managed_threads_lock:
            if name in self.__managed_threads_by_scope[scope_name]:
                # logger.error(f"A thread with name '{name}' is already registered, it will be overwritten: {self.__managed_threads_by_scope[name]}")
                logger.error(f"A thread with name '{name}' is already registered in scope '{scope_name}', it will be overwritten: {self.__managed_threads_by_scope[scope_name][name]}", stack_info=True)
            else:
                logger.debug(f"Registered thread with name '{name}' in scope '{scope_name}': {thread}")
            self.__managed_threads_by_scope[scope_name][name] = thread
        return scope_name

    def unregister_thread(self, name, scope=default_context):
        scope_name = self.__get_scope_name(scope)
        with self.__managed_threads_lock:
            if name in self.__managed_threads_by_scope[scope_name]:
                if self.__managed_threads_by_scope[scope_name][name].is_alive():
                    logger.warn(f"Unregistering thread with name '{name}' from scope '{scope_name}' whereas it is still alive")
                else:
                    logger.debug(f"Unregistered thread with name '{name}' from scope '{scope_name}'")
                del self.__managed_threads_by_scope[scope_name][name]
            else:
                raise Exception(f"Not registered thread of name '{name}' in scope '{scope_name}'")
        return scope_name

    def get_thread(self, name, scope=default_context):
        scope_name = self.__get_scope_name(scope)
        with self.__managed_threads_lock:
            if name in self.__managed_threads_by_scope[scope_name]:
                return self.__managed_threads_by_scope[scope_name][name]
            else:
                raise Exception(f"Not registered thread of name '{name}' in scope '{scope_name}'")

    def interrupt_thread(self, name, scope=default_context, with_join=True, join_timeout=undefined_argument, raise_if_not_interruptable=True):
        interrupt_status, join_status = False, False
        
        thread = self.get_thread(name, scope=scope)
        if thread.is_alive():
            if self.is_interruptable(thread=thread):
                try:
                    thread.interrupt()
                except Exception as exc:
                    # logger.exception("Failed to interrupt thread of name '{}'".format(name))
                    raise exc
                else:
                    interrupt_status = True
                
                if with_join:
                    join_status = thread.join(join_timeout)
            elif raise_if_not_interruptable:
                raise TechnicalException("Thread '{}' is not interruptable".format(name))
        
        return interrupt_status, join_status
        
    def join_thread(self, name, scope=default_context, timeout=undefined_argument):
        thread = self.get_thread(name, scope=scope)
        if thread.is_alive():
            thread.join(timeout=timeout)

    def interrupt_all_threads(self, scope=default_context, with_join=True, join_timeout=undefined_argument):
        """Interrupt all threads that are interruptable."""
        names = self.get_thread_names(scope=scope)
        logger.debug(f"Interrupting all threads in scope '{self.__get_scope_name(scope)}': {names}")
        interrupted_names = []
        for name in names:
            try:
                interrupt_status, _ = self.interrupt_thread(name, scope=scope, with_join=False, raise_if_not_interruptable=False)
            except Exception:
                logger.exception("Failed to interrupt thread of name '{}'".format(name))
            else:
                if interrupt_status:
                    interrupted_names.append(name)
        
        if with_join:
            self.__join_threads(interrupted_names, scope=scope, timeout=join_timeout, skip_not_interruptable=True)

    def join_all_threads(self, scope=default_context, timeout=undefined_argument, skip_not_interruptable=False):
        names = self.get_thread_names(scope=scope)
        self.__join_threads(names, scope=scope, timeout=timeout, skip_not_interruptable=skip_not_interruptable)

    def __join_threads(self, thread_names, scope=default_context, timeout=undefined_argument, skip_not_interruptable=False):
        for name in thread_names:
            if skip_not_interruptable and not self.is_interruptable(name=name, scope=scope):
                continue
            self.join_thread(name, scope=scope, timeout=timeout)
    
    def unregister_all_threads(self, scope=default_context, keep_alive=True):
        names = self.get_thread_names(scope=scope)
        for name in names:
            thread = self.get_thread(name, scope=scope)
            if not keep_alive or keep_alive and not thread.is_alive():
                self.unregister_thread(name, scope=scope)
    
    def get_thread_names(self, scope=default_context):
        scope_name = self.__get_scope_name(scope)
        with self.__managed_threads_lock:
            return sorted(list( self.__managed_threads_by_scope[scope_name].keys() ))
    
    def get_alive_thread_names(self, scope=default_context):
        scope_name = self.__get_scope_name(scope)
        with self.__managed_threads_lock:
            res = sorted(list( self.__managed_threads_by_scope[scope_name].keys() ))
            res = [name for name in res if self.__managed_threads_by_scope[scope_name][name].is_alive()]
            
    def is_interruptable(self, thread=None, name=None, scope=default_context):
        if name is not None:
            thread = self.get_thread(name, scope=scope)
        if thread is None:
            raise TechnicalException("At least one argument must be not None")
        
        return isinstance(thread, hmm.thread.InterruptableThread) and thread.is_interruptable
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from behave import given, when, then, step, Given, When, Then, Step  # @UnusedImport @UnresolvedImport
import behave
import re
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_core.common.tools.tools import Tools
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.common.context.session_context import SessionContext
from behave.model_core import Status
from behave.model_describe import escape_cell
from behave.textutil import indent
from holado_core.common.exceptions.holado_exception import HAException
from holado_test.scenario.step_tools import SRE  # @UnusedImport
from holado_multitask.multitasking.multitask_manager import MultitaskManager
from holado_python.standard_library.typing import Typing
from holado import is_in_steps_catalog

# limit import *
# pylint: disable=undefined-all-variable
__all__ = [
    "Given", "When", "Then", "Step",# "For",
    "format_step_with_context", "execute_steps", "SRE"
]

# Import and expose step parameter types
# from BehaveStepTools import SRE
# var_names = BehaveStepTools.import_types(sys.modules[__name__])
# if len(var_names) == 0:
#     raise TechnicalException(f"Module '{__name__}' is imported before registration of step parameter types")
# __all__.extend(var_names)


logger = logging.getLogger(__name__)

is_steps_catalog = is_in_steps_catalog()


def __get_report_manager():
    return SessionContext.instance().report_manager

def __has_scenario_context():
    return SessionContext.instance().has_scenario_context()

def __get_block_manager():
    if __has_scenario_context():
        return SessionContext.instance().get_scenario_context().block_manager
    else:
        return None

def __get_scope_manager():
    if __has_scenario_context():
        return SessionContext.instance().get_scenario_context().scope_manager
    else:
        return None




def format_step(step, with_keyword=True, with_table=True, with_text=True):
    from behave.model_describe import ModelDescriptor
    
    step_str = step.name.strip()
    
    if with_keyword:
        res = u"{keyword} {step_str}".format(keyword=step.keyword, step_str=step_str)
    else:
        res = step_str
        
    if with_table and hasattr(step, "table") and step.table is not None:
        rendered_table = render_step_table(step.table, "    ")
        res = u"{res}\n{table}".format(res=res, table=rendered_table)
    elif with_text and hasattr(step, "text") and step.text is not None:
        rendered_text = ModelDescriptor.describe_docstring(step.text, "")
        # rendered_text = Tools.indent_string(4, rendered_text)
        res = u"{res}\n{text}".format(res=res, text=rendered_text)
        
    return res

def format_step_with_context(context, step_str, with_keyword=True, with_table=True, with_text=True):
    from behave.model_describe import ModelDescriptor
    
    step_str = step_str.strip()
    
    if with_keyword:
        res = u"{keyword} {step_str}".format(keyword=context.step.keyword, step_str=step_str)
    else:
        res = step_str
        
    if with_table and hasattr(context, "table") and context.table is not None:
        rendered_table = render_step_table(context.table, "    ")
        res = u"{res}\n{table}".format(res=res, table=rendered_table)
    elif with_text and hasattr(context, "text") and context.text is not None:
        rendered_text = ModelDescriptor.describe_docstring(context.text, "    ")
        res = u"{res}\n{text}".format(res=res, text=rendered_text)
        
    return res

def __step_description(context):
    return __get_report_manager().StepTools.get_step_description(context.step)


def execute_steps(steps_text):
    """
    Wrapper to Context.execute_steps method, that manages scope levels of sub-steps.
    
    WARNING: The method context.execute_steps is not thread safe. 
             If it is called a second time in another thread very closely, the steps behind the second call
             will be executed two times, one time in each thread.
    TODO EKL: Do tests to make this function thread safe.
    """
    # Manage sub-step level
    thread_id = MultitaskManager.get_thread_id()
    scope_manager = __get_scope_manager()
    if scope_manager:
        level_already_exists = scope_manager.has_scope_level("steps")
        origin_level = scope_manager.scope_level("steps")
        if level_already_exists:
            scope_manager.increase_scope_level("steps")
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Executing steps (level {scope_manager.scope_level('steps')} ; thread: {thread_id}):\n{steps_text}")
    else:
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Executing steps (thread: {thread_id}):\n{steps_text}")
        
    try:
        SessionContext.instance().behave_manager.get_context().execute_steps(steps_text)
    finally:
        if scope_manager:
            scope_manager.reset_scope_level("steps", origin_level)

def render_step_table(table, indentation=None, do_escape_cell=False):
    """
    NOTE: It's a copy of behave method ModelDescriptor.describe_table, but not escaping cells by default.
          If parameter do_escape_cell is True, the behavior is identical to ModelDescriptor.describe_table.
    
    Provide a textual description of the table (as used w/ Gherkin).

    :param table:  Table to use (as :class:`behave.model.Table`)
    :param indentation:  Line prefix to use (as string, if any).
    :return: Textual table description (as unicode string).
    """
    # -- STEP: Determine output size of all cells.
    cell_lengths = []
    all_rows = [table.headings] + table.rows
    for row in all_rows:
        if do_escape_cell:
            lengths = [len(escape_cell(c)) for c in row]
        else:
            lengths = [len(c) for c in row]
        cell_lengths.append(lengths)

    # -- STEP: Determine max. output size for each column.
    max_lengths = []
    for col in range(0, len(cell_lengths[0])):
        max_lengths.append(max([c[col] for c in cell_lengths]))

    # -- STEP: Build textual table description.
    lines = []
    for r, row in enumerate(all_rows):
        line = u"|"
        for c, (cell, max_length) in enumerate(zip(row, max_lengths)):
            pad_size = max_length - cell_lengths[r][c]
            if do_escape_cell:
                line += u" %s%s |" % (escape_cell(cell), " " * pad_size)
            else:
                line += u" %s%s |" % (cell, " " * pad_size)
        line += u"\n"
        lines.append(line)

    if indentation:
        return indent(lines, indentation)
    # -- OTHERWISE:
    return u"".join(lines)


def Given(regexp, accept_scope_in_define=True, accept_condition=True):
    return _wrapped_step(behave.given, regexp, accept_scope_in_define=accept_scope_in_define, accept_condition=accept_condition, do_format_regexp=not is_steps_catalog)  #pylint: disable=no-member @UndefinedVariable

def Then(regexp):
    return _wrapped_step(behave.then, regexp, do_format_regexp=not is_steps_catalog)  #pylint: disable=no-member @UndefinedVariable

def When(regexp):
    return _wrapped_step(behave.when, regexp, do_format_regexp=not is_steps_catalog) #pylint: disable=no-member @UndefinedVariable

def Step(regexp):
    return _wrapped_step(behave.step, regexp, do_format_regexp=not is_steps_catalog) #pylint: disable=no-member @UndefinedVariable

# def For(regexp):
#     return _wrapped_step(behave.step, regexp, accept_expected_exception=False) #pylint: disable=no-member @UndefinedVariable


def _wrapped_step(step_function, regexp, accept_expected_exception=True, accept_condition=True, accept_scope_in_define=True, do_format_regexp=True):
    # Replace SRE types by associated regex
    if do_format_regexp:
        try:
            regexp = regexp.format(**SRE.all_as_dict())
        except IndexError:
            # This error occurs when regexp has no {} section to format
            pass
        except KeyError:
            # This error can occur when format is already done in file defining step
            pass
    
    def wrapper(func):
        res_func = func
        if accept_expected_exception:
            res_func = _accept_expected_exception(res_func)
        if accept_condition:
            res_func = _accept_condition(res_func)
        if accept_scope_in_define:
            res_func = _accept_scope_in_define(res_func)
        return step_function(regexp)(res_func)
    return wrapper

def _accept_condition(func):
    """
    If a condition is ongoing, and current condition is not valid, bypass the step execution.
    """
    def wrapper(context, *args, **kwargs):
        # logger.debug(f"++++++++++ Wrapper begin ; func={Typing.get_object_class_fullname(func)} ; context={context.__dict__} ; args={args} ; kwargs={kwargs}")
        if __get_scope_manager().is_in_condition() and not __get_scope_manager().is_in_valid_condition():
            step_context = __get_scope_manager().get_step_context()
            step_context.status = "skipped"
        else:
            func(context, *args, **kwargs)
        
    return wrapper

def _accept_scope_in_define(func):
    """
    If a block is under definition, bypass step execution.
    Note: in before_step, the step was added in scope under define.
    """
    def wrapper(context, *args, **kwargs):
        # logger.debug(f"++++++++++ Wrapper begin ; func={Typing.get_object_class_fullname(func)} ; context={context.__dict__} ; args={args} ; kwargs={kwargs}")
        if __get_block_manager().is_in_define:
            # Bypass step execution: in before_step, the step was added in scope under define
            pass
        else:
            func(context, *args, **kwargs)
        
    return wrapper



###################################################################################################
##  Manage expected exception

def _accept_expected_exception(func):
    """
    If an error is expected, check if an error really occurs and if it matches expected error.
    Otherwise raise it again.
    """
    def wrapper(context, *args, **kwargs):
        # logger.debug(f"++++++++++ Wrapper begin ; context={context.__dict__} ; args={args} ; kwargs={kwargs}")
        try:
            func(context, *args, **kwargs)
        except Exception as exc:  #pylint: disable=W0703
            __process_step_exception(context, exc)
        
        __process_expected_exception_at_end_of_step(context)
        
    return wrapper

def __process_step_exception(context, exc):
    if not __has_scenario_context():
        raise exc
    
    # logger.debug(f"++++++++++ Wrapper: exception catched: {exc}")
    exc_verified = False
    # Verify exception if an exception was expected
    if not context.in_expected_exception_step \
            and __get_scope_manager().scope_level("steps") == context.expected_exception_step_level \
            and MultitaskManager.get_thread_id() == context.expected_exception_thread_id:
        # logger.debug(f"++++++++++ Wrapper: verify catched exception")
        exc_verified = True
        if isinstance(exc, HAException) or not hasattr(context, "sub_step_exception") or context.sub_step_exception is None:
            exc_to_verify = exc
        else:
            exc_to_verify = context.sub_step_exception

        # Reset expected exception, only try matching once.
        expected_exception_str, expected_exception_pattern = pop_expected_exception(context)
        
        # Process verification
        # logger.debug(f"++++++++++ Wrapper: process verification on exception: {exc_to_verify}")
        if expected_exception_str:
            exc_str = "{}({})".format(Typing.get_object_class_fullname(exc_to_verify), str(exc_to_verify))
            if expected_exception_str != exc_str:
                raise VerifyException(f"Obtained exception is not the expected one:\n  obtained: {exc_str}\n  expected: {expected_exception_str}")
        elif expected_exception_pattern:
            exc_str = "{}({})".format(Typing.get_object_class_fullname(exc_to_verify), str(exc_to_verify))
            regex = re.compile(expected_exception_pattern, re.DOTALL)
            
            if not regex.search(exc_str):
                raise VerifyException(f"Obtained exception is not matching the expected pattern:\n  obtained: {exc_str}\n  expected pattern: '{expected_exception_pattern}'")
        
        # Update status
        # logger.debug(f"++++++++++ Wrapper: update behave objects")
        context.do_update_step_status = Status.passed
    
    # Store as sub-step exception
    if __get_scope_manager().scope_level("steps") > 0 and not exc_verified:
        context.sub_step_exception = exc

    # If no other exception is raised, re-raise the same exception
    if not exc_verified:
        # logger.debug(f"++++++++++ Wrapper: raise catched exception")
        raise

def __process_expected_exception_at_end_of_step(context):
    if context.in_expected_exception_step:
        # logger.debug(f"++++++++++ Wrapper: Finish declaration of step setting expected exception")
        # Finish declaration of step setting expected exception
        leave_expected_exception_step(context)
    elif __get_scope_manager().scope_level("steps") == context.expected_exception_step_level \
            and MultitaskManager.get_thread_id() == context.expected_exception_thread_id:
        # logger.debug(f"++++++++++ Wrapper: Manage expected exception has not occurred")
        # Manage expected exception has not occurred
        if context.expected_exception_str:
            # Raise exception to make scenario failing, since step hasn't failed as expected
            msg = "Step passed but was expected fail on an exception:\n    step:\n{}\n    expected exception:\n{}".format(
                Tools.indent_string(8, __step_description(context)), 
                Tools.indent_string(8, context.expected_exception_str))
            if context.step.hook_failed:
                msg += "\n    Hook failed"
            logger.error(msg)
            raise FunctionalException(msg)
        elif context.expected_exception_pattern:
            # Raise exception to make scenario failing, since step hasn't failed as expected
            msg = "Step passed but was expected fail on an exception:\n    step:\n{}\n    expected exception matching: '{}'".format(
                Tools.indent_string(8, __step_description(context)), 
                Tools.indent_string(8, context.expected_exception_pattern))
            if context.step.hook_failed:
                msg += "\n    Hook failed"
            logger.error(msg)
            raise FunctionalException(msg)
        else:
            raise TechnicalException("No expected exception is configured")
            
def enter_expected_exception_step(context, expected_exception_str = None, expected_exception_pattern=None):
    context.expected_exception_str = expected_exception_str
    context.expected_exception_pattern = expected_exception_pattern
    context.in_expected_exception_step = True
    context.expected_exception_step_level = __get_scope_manager().scope_level("steps")
    context.expected_exception_thread_id = MultitaskManager.get_thread_id()

def leave_expected_exception_step(context):
    context.in_expected_exception_step = False

def pop_expected_exception(context):
    expected_exception_str = context.expected_exception_str
    expected_exception_pattern = context.expected_exception_pattern
    context.expected_exception_str = None
    context.expected_exception_pattern = None
    context.expected_exception_step_level = -1
    context.expected_exception_thread_id = None
    return expected_exception_str, expected_exception_pattern



###################################################################################################
##  Setup custom step decorators
###################################################################################################

### /!\ Not yet working, that's why next lines are commented
### Next lines enable the decorator, so that declaration of steps with this decorator is working.
### But behave has to be patched to parse .feature correctly and considering these decorators as steps:
###   - in behave/parser.py: in Parser.parse_step, manage the new step types ('for')
###   - in behave/i18n.py: add the new step types ('for')  

# def setup_step_decorators():
#     registry=behave.step_registry.registry
#     for step_type in ("for",):
#         registry.steps[step_type] = []
#         registry.make_decorator(step_type)
#
# setup_step_decorators()





# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_test.behave.behave import render_step_table
import logging
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_system.system.filesystem.file import File
import base64
import csv
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.tables.table_row import TableRow
import os.path
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.standard_library.csv import CsvManager

logger = logging.getLogger(__name__)



def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()

def __get_report_manager():
    return SessionContext.instance().report_manager

def __get_path_manager():
    return SessionContext.instance().path_manager


@Given(r"(?P<var_name>{Variable}) = table with CSV content (?P<content>{Str})")
def step_impl(context, var_name, content):
    var_name = StepTools.evaluate_variable_name(var_name)
    execute_steps(u"""
        Given __FILE_CSV__@ = create CSV file with prefix 'csv_content' and content {content}
        Given {var_name} = table with content of CSV file __FILE_CSV__@
        """.format(var_name=var_name, content=content) )

@Given(r"(?P<var_name>{Variable}) = table with content of CSV file (?P<path>{Str})(?: \((?:encoding: (?P<encoding>{Str}))?(?: ; )?(?:dialect: (?P<dialect>{Str}))?(?: ; )?(?:delimiter: (?P<delimiter>{Str}))?(?: ; )?(?:quotechar: (?P<quotechar>{Str}))?(?: ; )?(?:quoting: (?P<quoting_str>ALL|MINIMAL|NONE|NONNUMERIC))?\))?")
def step_impl(context, var_name, path, encoding, dialect, delimiter, quotechar, quoting_str):
    var_name = StepTools.evaluate_variable_name(var_name)
    file_path = StepTools.evaluate_scenario_parameter(path)
    encoding = StepTools.evaluate_scenario_parameter(encoding)
    dialect = StepTools.evaluate_scenario_parameter(dialect)
    delimiter = StepTools.evaluate_scenario_parameter(delimiter)
    quotechar = StepTools.evaluate_scenario_parameter(quotechar)
    quoting_str = StepTools.evaluate_scenario_parameter(quoting_str)
    quoting = getattr(csv, f"QUOTE_{quoting_str}") if quoting_str is not None else None
    
    res = CsvManager.table_with_content_of_CSV_file(file_path, encoding, dialect, delimiter, quotechar, quoting)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = create CSV file with prefix (?P<path_prefix>{Str})")
def step_impl(context, var_name, path_prefix):
    var_name = StepTools.evaluate_variable_name(var_name)
    rendered_table = render_step_table(context.table, "    ")

    execute_steps(u"""
        Given __FILENAME__@ = path to timestamped file with prefix {path_prefix} and extension 'csv'
        Given {var_name} = create CSV file with name __FILENAME__@
        {table}
        """.format(path_prefix=path_prefix, var_name=var_name, table=rendered_table))

@Given(r"(?P<var_name>{Variable}) = create CSV file with name (?P<filename>{Str})")
def step_impl(context, var_name, filename):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = BehaveStepTools.get_step_table(context)
    
    __get_variable_manager().register_variable("__TABLE_CONTENT__@", table)
    execute_steps(u"""
        Given {var_name} = create CSV file with name {filename} and table content __TABLE_CONTENT__@
        """.format(var_name=var_name, filename=filename) )

@Given(r"(?P<var_name>{Variable}) = create CSV file with name (?P<filename>{Str}) and table content (?P<table_content>{Variable})")
def step_impl(context, var_name, filename, table_content):
    var_name = StepTools.evaluate_variable_name(var_name)
    filename = StepTools.evaluate_scenario_parameter(filename)
    
    dest_path = __get_report_manager().current_scenario_report.get_path("files")
    file_path = os.path.join(dest_path, filename)
    __get_variable_manager().register_variable("__FILE_PATH__@", file_path)
    
    execute_steps(u"""
        Given create CSV file with path __FILE_PATH__@ and table content {table_content}
        """.format(table_content=table_content) )

    __get_path_manager().makedirs(file_path)
    
    __get_variable_manager().register_variable(var_name, file_path)

@Given(r"create CSV file with path (?P<file_path>{Str}) and table content (?P<table_content>{Variable})(?: \((?:encoding: (?P<encoding>{Str}))?(?: ; )?(?:dialect: (?P<dialect>{Str}))?(?: ; )?(?:delimiter: (?P<delimiter>{Str}))?(?: ; )?(?:quotechar: (?P<quotechar>{Str}))?(?: ; )?(?:quoting: (?P<quoting_str>ALL|MINIMAL|NONE|NONNUMERIC))?\))?")
def step_impl(context, file_path, table_content, encoding, dialect, delimiter, quotechar, quoting_str):
    file_path = StepTools.evaluate_scenario_parameter(file_path)
    table = StepTools.evaluate_variable_value(table_content)
    encoding = StepTools.evaluate_scenario_parameter(encoding)
    dialect = StepTools.evaluate_scenario_parameter(dialect)
    delimiter = StepTools.evaluate_scenario_parameter(delimiter)
    quotechar = StepTools.evaluate_scenario_parameter(quotechar)
    quoting_str = StepTools.evaluate_scenario_parameter(quoting_str)
    quoting = getattr(csv, f"QUOTE_{quoting_str}") if quoting_str is not None else None
    
    CsvManager.create_csv_file(file_path, table, encoding, dialect, delimiter, quotechar, quoting)

@Given(r"(?P<var_name>{Variable}) = create CSV file with prefix (?P<path_prefix>{Str}) and content (?P<content>{Str})")
def step_impl(context, var_name, path_prefix, content):
    var_name = StepTools.evaluate_variable_name(var_name)
    execute_steps(u"""
        Given {var_name} = create file with prefix {path_prefix} and extension 'csv' and content {content}
        """.format(var_name=var_name, path_prefix=path_prefix, content=content) )



